"""
    Copyright 2022 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
import asyncio
import dataclasses
import datetime
import logging
import uuid
from concurrent.futures import ThreadPoolExecutor
from functools import partial
from queue import Queue
from threading import Event
from typing import Optional

import pytest

import inmanta
from inmanta import util
from inmanta.util import (
    CronSchedule,
    CycleException,
    IntervalSchedule,
    NamedLock,
    ScheduledTask,
    TaskSchedule,
    ensure_future_and_handle_exception,
    join_threadpools,
    stable_depth_first,
)
from utils import LogSequence, get_product_meta_data, log_contains, no_error_in_logs

LOGGER = logging.getLogger(__name__)


def test_interval_schedule() -> None:
    """
    Verifies that the IntervalSchedule class' primary methods work as expected.
    """
    simple: TaskSchedule = IntervalSchedule(interval=1.0)
    assert simple.get_initial_delay() == 1.0
    assert simple.get_next_delay() == 1.0
    with_delay: TaskSchedule = IntervalSchedule(interval=1.0, initial_delay=0.0)
    assert with_delay.get_initial_delay() == 0.0
    assert with_delay.get_next_delay() == 1.0


def test_interval_schedule_equals() -> None:
    """
    Verifies that the IntervalSchedule class' equality checks work as expected.
    """
    assert IntervalSchedule(interval=1.0) == IntervalSchedule(interval=1.0)
    assert IntervalSchedule(interval=1.0) != IntervalSchedule(interval=2.0)
    assert IntervalSchedule(interval=1.0, initial_delay=0.5) == IntervalSchedule(interval=1.0, initial_delay=0.5)
    assert IntervalSchedule(interval=1.0, initial_delay=0.5) != IntervalSchedule(interval=1.0, initial_delay=0.0)


def test_cron_schedule(monkeypatch) -> None:
    """
    Verifies that the CronSchedule class' primary methods work as expected.
    """

    def freeze_time(time: datetime.datetime) -> None:
        """
        Freeze time to a given value to avoid race conditions.
        """

        class frozendatetime(datetime.datetime):
            @classmethod
            def now(cls, tz: Optional[datetime.tzinfo] = None) -> datetime.datetime:
                if tz is None:
                    return time if time.tzinfo is None else time.astimezone(datetime.timezone.utc).replace(tzinfo=None)
                else:
                    aware: datetime.datetime = time if time.tzinfo is not None else time.replace(tzinfo=datetime.timezone.utc)
                    return aware.astimezone(tz)

        monkeypatch.setattr(datetime, "datetime", frozendatetime)

    # min hour day month dow
    schedule: TaskSchedule = CronSchedule(cron="1 2 * * *")

    # year month day hour min sec ms
    now: datetime.datetime = datetime.datetime(2022, 6, 16, 10, 15, 3, 0)
    freeze_time(now)
    next_delta: datetime.timedelta = datetime.datetime(2022, 6, 17, 2, 1, 0, 0) - now
    assert schedule.get_initial_delay() == schedule.get_next_delay() == next_delta.total_seconds()

    tomorrow: datetime.datetime = now + datetime.timedelta(days=1, seconds=42)
    freeze_time(tomorrow)
    assert schedule.get_initial_delay() == schedule.get_next_delay() == next_delta.total_seconds() - 42


def test_cron_schedule_equals() -> None:
    """
    Verifies that the CronSchedule class' equality checks work as expected.
    """
    assert CronSchedule(cron="1 2 * * *") == CronSchedule(cron="1 2 * * *")
    assert CronSchedule(cron="1 2 * * *") != CronSchedule(cron="0 2 * * *")


async def test_scheduler_remove(caplog):
    scheduler = util.Scheduler("remove")

    i = []

    async def action():
        i.append(0)

    schedule: IntervalSchedule = IntervalSchedule(0.05, 0)
    scheduler.add_action(action, schedule)

    while len(i) == 0:
        await asyncio.sleep(0.01)

    assert ScheduledTask(action=action, schedule=schedule) in scheduler._scheduled
    scheduler.remove(ScheduledTask(action=action, schedule=schedule))
    length = len(i)
    await asyncio.sleep(0.1)
    assert len(i) == length
    assert not scheduler._executing_tasks[action]
    assert not (action, schedule) in scheduler._scheduled
    no_error_in_logs(caplog)


async def test_scheduler_remove_same_action() -> None:
    """
    Verify that removing an action from the scheduler works as intended in the presence of other similar scheduled actions.
    As long as there exists no other scheduled action for exactly the same action and schedule, removal should be unambiguous.
    This test includes verification of potential edge cases such as `partial` of the same function, locally defined functions
    within a body that gets evaluated twice, ...
    """

    scheduler = util.Scheduler("remove_same_action")

    async def myaction() -> None:
        pass

    schedule_one: TaskSchedule = IntervalSchedule(100)
    schedule_two: TaskSchedule = IntervalSchedule(200)

    # same action, different schedule => expect distinct tasks
    scheduler.add_action(myaction, schedule_one)
    scheduler.add_action(myaction, schedule_two)

    # same action wrapped in `partial`, same schedule => expect distinct tasks
    partial_one = partial(myaction)
    partial_two = partial(myaction)
    assert partial_one != partial_two
    scheduler.add_action(partial_one, schedule_one)
    scheduler.add_action(partial_two, schedule_one)

    locally_defined: list[object] = []

    def add_locally_defined() -> None:
        async def myaction() -> None:
            pass

        assert ScheduledTask(action=myaction, schedule=schedule_one) not in scheduler._scheduled
        scheduler.add_action(myaction, schedule_one)
        locally_defined.append(myaction)

    add_locally_defined()
    add_locally_defined()

    scheduler.add_action(myaction, schedule_two)
    assert ScheduledTask(action=myaction, schedule=schedule_one) in scheduler._scheduled
    assert ScheduledTask(action=myaction, schedule=schedule_two) in scheduler._scheduled
    assert ScheduledTask(action=partial_one, schedule=schedule_one) in scheduler._scheduled
    assert ScheduledTask(action=partial_two, schedule=schedule_one) in scheduler._scheduled
    assert ScheduledTask(action=locally_defined[0], schedule=schedule_one) in scheduler._scheduled
    assert ScheduledTask(action=locally_defined[1], schedule=schedule_one) in scheduler._scheduled

    scheduler.remove(ScheduledTask(action=myaction, schedule=schedule_one))
    assert ScheduledTask(action=myaction, schedule=schedule_one) not in scheduler._scheduled
    assert ScheduledTask(action=myaction, schedule=schedule_two) in scheduler._scheduled
    assert ScheduledTask(action=partial_one, schedule=schedule_one) in scheduler._scheduled
    assert ScheduledTask(action=partial_two, schedule=schedule_one) in scheduler._scheduled
    assert ScheduledTask(action=locally_defined[0], schedule=schedule_one) in scheduler._scheduled
    assert ScheduledTask(action=locally_defined[1], schedule=schedule_one) in scheduler._scheduled

    scheduler.remove(ScheduledTask(action=myaction, schedule=schedule_two))
    assert ScheduledTask(action=myaction, schedule=schedule_two) not in scheduler._scheduled
    assert ScheduledTask(action=partial_one, schedule=schedule_one) in scheduler._scheduled
    assert ScheduledTask(action=partial_two, schedule=schedule_one) in scheduler._scheduled
    assert ScheduledTask(action=locally_defined[0], schedule=schedule_one) in scheduler._scheduled
    assert ScheduledTask(action=locally_defined[1], schedule=schedule_one) in scheduler._scheduled

    scheduler.remove(ScheduledTask(action=partial_one, schedule=schedule_one))
    assert ScheduledTask(action=partial_one, schedule=schedule_one) not in scheduler._scheduled
    assert ScheduledTask(action=partial_two, schedule=schedule_one) in scheduler._scheduled
    assert ScheduledTask(action=locally_defined[0], schedule=schedule_one) in scheduler._scheduled
    assert ScheduledTask(action=locally_defined[1], schedule=schedule_one) in scheduler._scheduled

    scheduler.remove(ScheduledTask(action=partial_two, schedule=schedule_one))
    assert ScheduledTask(action=partial_two, schedule=schedule_one) not in scheduler._scheduled
    assert ScheduledTask(action=locally_defined[0], schedule=schedule_one) in scheduler._scheduled
    assert ScheduledTask(action=locally_defined[1], schedule=schedule_one) in scheduler._scheduled

    scheduler.remove(ScheduledTask(action=locally_defined[0], schedule=schedule_one))
    assert ScheduledTask(action=locally_defined[0], schedule=schedule_one) not in scheduler._scheduled
    assert ScheduledTask(action=locally_defined[1], schedule=schedule_one) in scheduler._scheduled

    scheduler.remove(ScheduledTask(action=locally_defined[1], schedule=schedule_one))
    assert ScheduledTask(action=locally_defined[1], schedule=schedule_one) not in scheduler._scheduled


async def test_scheduler_stop(caplog):
    sched = util.Scheduler("stop")

    i = []

    async def action():
        i.append(0)
        return "A"

    sched.add_action(action, IntervalSchedule(0.05, 0))

    while len(i) == 0:
        await asyncio.sleep(0.01)

    await sched.stop()

    length = len(i)
    await asyncio.sleep(0.1)
    assert len(i) == length
    no_error_in_logs(caplog)

    caplog.clear()
    sched.add_action(action, IntervalSchedule(0.05, 0))
    assert "Scheduling action 'action', while scheduler is stopped" in caplog.messages
    assert not sched._executing_tasks[action]


async def test_scheduler_async_run_fail(caplog):
    sched = util.Scheduler("xxx")

    i = []

    async def action():
        i.append(0)
        await asyncio.sleep(0)
        raise Exception("Marker")

    sched.add_action(action, IntervalSchedule(0.05, 0))

    while len(i) == 0:
        await asyncio.sleep(0.01)

    await sched.stop()

    length = len(i)
    await asyncio.sleep(0.1)
    assert len(i) == length
    assert not sched._executing_tasks[action]

    print(caplog.messages)

    log_contains(caplog, "inmanta.util", logging.ERROR, "Uncaught exception while executing scheduled action")


async def test_scheduler_run_async(caplog):
    sched = util.Scheduler("xxx")

    i = []

    async def action():
        i.append(0)

    sched.add_action(action, IntervalSchedule(0.05, 0))

    while len(i) == 0:
        await asyncio.sleep(0.01)

    await sched.stop()

    length = len(i)
    await asyncio.sleep(0.1)
    assert len(i) == length
    assert not sched._executing_tasks[action]
    no_error_in_logs(caplog)


async def test_scheduler_cancel_executing_tasks() -> None:
    """
    Verify that executing tasks are cancelled when the scheduler is stopped.
    """

    @dataclasses.dataclass
    class TaskStatus:
        task_is_executing: bool = False
        task_was_cancelled: bool = False

    task_status = TaskStatus()

    async def action():
        task_status.task_is_executing = True
        try:
            await asyncio.sleep(1000)
        except asyncio.CancelledError:
            task_status.task_was_cancelled = True
            raise

    sched = util.Scheduler("xxx")
    sched.add_action(action, IntervalSchedule(interval=1000, initial_delay=0))
    await util.retry_limited(lambda: task_status.task_is_executing, timeout=10)
    assert task_status.task_is_executing
    assert not task_status.task_was_cancelled
    assert sched._executing_tasks[action]
    await sched.stop()
    await util.retry_limited(lambda: task_status.task_was_cancelled, timeout=10)
    assert not sched._executing_tasks[action]


@pytest.mark.parametrize("cancel_on_stop", [True, False])
async def test_scheduler_waits_on_shutdown(cancel_on_stop) -> None:
    """
    Verify that tasks can be tagged to be awaited when the scheduler is stopped.
    """

    @dataclasses.dataclass
    class TaskStatus:
        task_is_executing: bool = False
        task_was_cancelled: bool = False
        task_was_executed: bool = False

    task_status = TaskStatus()

    async def action():
        task_status.task_is_executing = True
        try:
            await asyncio.sleep(0.2)
        except asyncio.CancelledError:
            task_status.task_was_cancelled = True
            raise
        task_status.task_was_executed = True

    sched = util.Scheduler("test_await_tasks_on_shutdown")
    sched.add_action(action, IntervalSchedule(interval=1000, initial_delay=0), cancel_on_stop=cancel_on_stop)
    await util.retry_limited(lambda: task_status.task_is_executing, timeout=3)

    await sched.stop()

    if cancel_on_stop:
        await util.retry_limited(lambda: task_status.task_was_cancelled, timeout=3)
        assert not task_status.task_was_executed
    else:
        await util.retry_limited(lambda: task_status.task_was_executed, timeout=3)
        assert not task_status.task_was_cancelled


async def test_ensure_future_and_handle_exception(caplog):
    caplog.set_level(logging.INFO)

    async def success():
        LOGGER.info("Success")

    async def fail():
        LOGGER.info("Fail")
        raise Exception("message F")

    ensure_future_and_handle_exception(LOGGER, "marker 1", success(), notify_done_callback=lambda x: None)
    ensure_future_and_handle_exception(LOGGER, "marker 2", fail(), notify_done_callback=lambda x: None)

    await asyncio.sleep(0.2)

    LogSequence(caplog).contains("test_util", logging.INFO, "Success")
    final = (
        LogSequence(caplog).contains("test_util", logging.INFO, "Fail").contains("test_util", logging.ERROR, "marker 2").index
        - 1
    )
    exception = caplog.get_records("call")[final].exc_info[1]
    assert str(exception) == "message F"


def test_stable_dfs():
    def expand_graph(gs):
        """expand a graph od the form
        a: a b c d
        """
        nodes = set()
        edges = {}

        for line in gs.split("\n"):
            if not line.strip():
                continue
            f, t = line.split(":")
            f = f.strip()
            if not f:
                continue
            nodes.add(f)
            t = t.strip()
            if not t:
                continue
            ts = [target.strip() for target in t.split(" ") if target.strip()]
            for target in ts:
                nodes.add(target)
            edges[f] = ts
        return list(nodes), edges

    graph = expand_graph(
        """
    e: f
    a: b c
    b: c d
    h: i
    0:
    """
    )
    seq = stable_depth_first(*graph)
    assert seq == ["0", "c", "d", "b", "a", "f", "e", "i", "h"]

    graph = expand_graph(
        """
        e: f
        b: c d
        a: c b
        h: i
        0:
        """
    )
    seq = stable_depth_first(*graph)
    assert seq == ["0", "c", "d", "b", "a", "f", "e", "i", "h"]

    with pytest.raises(CycleException) as e:
        stable_depth_first(*expand_graph("a: a"))

    assert e.value.nodes == ["a"]

    with pytest.raises(CycleException) as e:
        stable_depth_first(
            *expand_graph(
                """a: b
        b: a"""
            )
        )

    assert e.value.nodes == ["b", "a"]

    # missing nodes
    graph, edges = expand_graph("""a: b""")
    graph.remove("b")

    seq = stable_depth_first(graph, edges)
    assert seq == ["b", "a"]


def test_is_sub_dict():
    identifier = uuid.uuid4()
    now = datetime.datetime.now()
    dct = {1: 2, "test": False, "date": now, "id": identifier, "str": "string"}

    assert util.is_sub_dict({}, dct)
    assert util.is_sub_dict({1: 2}, dct)
    assert util.is_sub_dict({"test": False}, dct)
    assert util.is_sub_dict({"date": now}, dct)
    assert util.is_sub_dict({"id": identifier}, dct)
    assert util.is_sub_dict({"str": "string"}, dct)
    assert util.is_sub_dict({"test": False, "date": now}, dct)
    assert not util.is_sub_dict({"test": True, "date": now}, dct)
    assert not util.is_sub_dict({"test": False, "date": datetime.datetime.now()}, dct)
    assert not util.is_sub_dict({1: 2, "test": False, "date": now, "id": identifier, "val": "val"}, dct)


def test_get_product_meta_data():
    """Basic smoke test for testing utils"""
    assert get_product_meta_data() is not None


async def test_named_lock():
    lock = NamedLock()
    await lock.acquire("a")
    await lock.acquire("b")
    await lock.release("b")
    fut = asyncio.create_task(lock.acquire("a"))
    assert not fut.done()
    await lock.release("a")
    await fut
    await lock.release("a")
    # Don't leak
    assert not lock._named_locks


def test_running_test_fixture():
    """
    Assert that the RUNNING_TESTS variable is set to True when we run the tests
    """
    assert inmanta.RUNNING_TESTS


async def test_threadpool_join():
    tp = ThreadPoolExecutor()

    hanglock = Event()
    done = Queue()

    eventloop = asyncio.get_event_loop()

    async def cor():
        await asyncio.sleep(0.02)

    def worker():
        # hang on lock
        hanglock.wait()
        # hang on ioloop
        block = asyncio.run_coroutine_threadsafe(cor(), loop=eventloop)
        block.result()
        done.put("A")

    for i in range(5):
        tp.submit(worker)

    tp.shutdown(wait=False)
    assert done.qsize() == 0
    hanglock.set()
    assert done.qsize() == 0
    await join_threadpools([tp])
    # verify we are done
    tp.shutdown(wait=True)
    assert done.qsize() == 5



"""
    Copyright 2019 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com

    Module defining the v1 rest api
"""

import datetime
import uuid
from typing import Any, Dict, List, Optional, Union

from inmanta import const, data, resources
from inmanta.data import model
from inmanta.types import JsonType, PrimitiveTypes

from . import exceptions
from .common import ArgOption
from .decorators import method, typedmethod


async def convert_environment(env: uuid.UUID, metadata: dict) -> "data.Environment":
    metadata[const.INMANTA_URN + "env"] = str(env)
    env = await data.Environment.get_by_id(env)
    if env is None:
        raise exceptions.NotFound("The given environment id does not exist!")
    return env


async def add_env(env: uuid.UUID, metadata: dict) -> uuid.UUID:
    metadata[const.INMANTA_URN + "env"] = str(env)
    return env


async def ignore_env(obj: Any, metadata: dict) -> Any:
    """
    This mapper only adds an env all for authz
    """
    metadata[const.INMANTA_URN + "env"] = "all"
    return obj


async def convert_resource_version_id(rvid: model.ResourceVersionIdStr, metadata: dict) -> "resources.Id":
    try:
        return resources.Id.parse_resource_version_id(rvid)
    except Exception:
        raise exceptions.BadRequest(f"Invalid resource version id: {rvid}")


ENV_OPTS: Dict[str, ArgOption] = {
    "tid": ArgOption(header=const.INMANTA_MT_HEADER, reply_header=True, getter=convert_environment)
}
AGENT_ENV_OPTS = {"tid": ArgOption(header=const.INMANTA_MT_HEADER, reply_header=True, getter=add_env)}
RVID_OPTS = {"rvid": ArgOption(getter=convert_resource_version_id)}


# Method for working with projects
@method(path="/project", operation="PUT", client_types=[const.ClientType.api])
def create_project(name: str, project_id: uuid.UUID = None):
    """
    Create a new project

    :param name: The name of the project
    :param project_id: A unique uuid, when it is not provided the server generates one
    """


@method(path="/project/<id>", operation="POST", client_types=[const.ClientType.api])
def modify_project(id: uuid.UUID, name: str):
    """
    Modify the given project
    """


@method(path="/project/<id>", operation="DELETE", client_types=[const.ClientType.api])
def delete_project(id: uuid.UUID):
    """
    Delete the given project and all related data
    """


@method(path="/project", operation="GET", client_types=[const.ClientType.api])
def list_projects():
    """
    Returns a list of projects ordered alphabetically by name. The environments within each project are also sorted by name.
    """


@method(path="/project/<id>", operation="GET", client_types=[const.ClientType.api])
def get_project(id: uuid.UUID):
    """
    Get a project and a list of the ids of all environments
    """


# Methods for working with environments
@method(path="/environment", operation="PUT", client_types=[const.ClientType.api])
def create_environment(
    project_id: uuid.UUID, name: str, repository: str = None, branch: str = None, environment_id: uuid.UUID = None
):
    """
    Create a new environment

    :param project_id: The id of the project this environment belongs to
    :param name: The name of the environment
    :param repository: The url (in git form) of the repository
    :param branch: The name of the branch in the repository
    :param environment_id: A unique environment id, if none an id is allocated by the server
    """


@method(path="/environment/<id>", operation="POST", client_types=[const.ClientType.api])
def modify_environment(id: uuid.UUID, name: str, repository: str = None, branch: str = None):
    """
    Modify the given environment

    :param id: The id of the environment
    :param name: The name of the environment
    :param repository: The url (in git form) of the repository
    :param branch: The name of the branch in the repository
    """


@method(path="/environment/<id>", operation="DELETE", client_types=[const.ClientType.api])
def delete_environment(id: uuid.UUID):
    """
    Delete the given environment and all related data.

    :param id: The uuid of the environment.

    :raises NotFound: The given environment doesn't exist.
    :raises Forbidden: The given environment is protected.
    """


@method(path="/environment", operation="GET", client_types=[const.ClientType.api])
def list_environments():
    """
    Returns a list of environments. The results are sorted by (project id, environment name, environment id).
    """


@method(
    path="/environment/<id>",
    operation="GET",
    client_types=[const.ClientType.api],
    arg_options={"id": ArgOption(getter=add_env)},
)
def get_environment(id: uuid.UUID, versions: int = None, resources: int = None):
    """
    Get an environment and all versions associated

    :param id: The id of the environment to return
    :param versions: Include this many available version for this environment.
    :param resources: Include this many available resources for this environment.
    """


# Method for listing/getting/setting/removing settings of an environment. This API is also used by agents to configure
# environments.


@method(
    path="/environment_settings",
    operation="GET",
    arg_options=ENV_OPTS,
    api=True,
    agent_server=True,
    client_types=[const.ClientType.api, const.ClientType.agent, const.ClientType.compiler],
)
def list_settings(tid: uuid.UUID):
    """
    List the settings in the current environment ordered by name alphabetically.
    """


@method(
    path="/environment_settings/<id>",
    operation="POST",
    arg_options=ENV_OPTS,
    api=True,
    agent_server=True,
    client_types=[const.ClientType.api, const.ClientType.agent, const.ClientType.compiler],
)
def set_setting(tid: uuid.UUID, id: str, value: Union[PrimitiveTypes, JsonType]):
    """
    Set a value
    """


@method(
    path="/environment_settings/<id>",
    operation="GET",
    arg_options=ENV_OPTS,
    api=True,
    agent_server=True,
    client_types=[const.ClientType.api, const.ClientType.agent],
)
def get_setting(tid: uuid.UUID, id: str):
    """
    Get a value
    """


@method(
    path="/environment_settings/<id>",
    operation="DELETE",
    arg_options=ENV_OPTS,
    api=True,
    agent_server=True,
    client_types=[const.ClientType.api, const.ClientType.agent],
)
def delete_setting(tid: uuid.UUID, id: str):
    """
    Delete a value
    """


# Method for listing and creating auth tokens for an environment that can be used by the agent and compilers


@method(
    path="/environment_auth",
    operation="POST",
    arg_options=ENV_OPTS,
    client_types=[const.ClientType.api, const.ClientType.compiler],
)
def create_token(tid: uuid.UUID, client_types: list, idempotent: bool = True):
    """
    Create or get a new token for the given client types. Tokens generated with this call are scoped to the current
    environment.

    :param tid: The environment id
    :param client_types: The client types for which this token is valid (api, agent, compiler)
    :param idempotent: The token should be idempotent, such tokens do not have an expire or issued at set so their
                       value will not change.
    """


#  Decomission an environment


@method(
    path="/decommission/<id>",
    operation="POST",
    arg_options={"id": ArgOption(getter=convert_environment)},
    client_types=[const.ClientType.api],
    api_version=1,
)
def decomission_environment(id: uuid.UUID, metadata: dict = None):
    """
    Decommision an environment. This is done by uploading an empty model to the server and let purge_on_delete handle
    removal.

    :param id: The uuid of the environment.
    :param metadata: Optional metadata associated with the decommissioning

    :raises NotFound: The given environment doesn't exist.
    :raises Forbidden: The given environment is protected.
    """


@method(
    path="/decommission/<id>",
    operation="DELETE",
    arg_options={"id": ArgOption(getter=convert_environment)},
    client_types=[const.ClientType.api],
)
def clear_environment(id: uuid.UUID):
    """
    Clear all data from this environment.

    :param id: The uuid of the environment.

    :raises NotFound: The given environment doesn't exist.
    :raises Forbidden: The given environment is protected.

    """


# Send a heartbeat to indicate that an agent is alive
@method(
    path="/heartbeat",
    operation="POST",
    agent_server=True,
    validate_sid=False,
    arg_options=ENV_OPTS,
    client_types=[const.ClientType.agent],
)
def heartbeat(sid: uuid.UUID, tid: uuid.UUID, endpoint_names: list, nodename: str, no_hang: bool = False):
    """
    Send a heartbeat to the server

    :param sid: The session ID used by this agent at this moment
    :param tid: The environment this node and its agents belongs to
    :param endpoint_names: The names of the endpoints on this node
    :param nodename: The name of the node from which the heart beat comes
    :param no_hang: don't use this call for long polling, but for connectivity check

    also registered as API method, because it is called with an invalid SID the first time
    """


@method(
    path="/heartbeat",
    operation="PUT",
    agent_server=True,
    arg_options={"sid": ArgOption(getter=ignore_env)},
    client_types=[const.ClientType.agent],
)
def heartbeat_reply(sid: uuid.UUID, reply_id: uuid.UUID, data: dict):
    """
    Send a reply back to the server

    :param sid: The session ID used by this agent at this moment
    :param reply_id: The id data is a reply to
    :param data: The data as a response to the reply
    """


# Upload, retrieve and check for file. A file is identified by a hash of its content.


@method(
    path="/file/<id>",
    operation="PUT",
    agent_server=True,
    api=True,
    client_types=[const.ClientType.api, const.ClientType.agent, const.ClientType.compiler],
    arg_options={"id": ArgOption(getter=ignore_env)},
)
def upload_file(id: str, content: str):
    """
    Upload a new file

    :param id: The id of the file
    :param content: The base64 encoded content of the file
    """


@method(
    path="/file/<id>",
    operation="HEAD",
    agent_server=True,
    api=True,
    client_types=[const.ClientType.api, const.ClientType.agent, const.ClientType.compiler],
    arg_options={"id": ArgOption(getter=ignore_env)},
)
def stat_file(id: str):
    """
    Does the file exist

    :param id: The id of the file to check
    """


@method(
    path="/file/<id>",
    operation="GET",
    agent_server=True,
    api=True,
    client_types=[const.ClientType.api, const.ClientType.agent, const.ClientType.compiler],
    arg_options={"id": ArgOption(getter=ignore_env)},
)
def get_file(id: str):
    """
    Retrieve a file

    :param id: The id of the file to retrieve
    """


@method(
    path="/file",
    api=True,
    client_types=[const.ClientType.api, const.ClientType.agent, const.ClientType.compiler],
    arg_options={"files": ArgOption(getter=ignore_env)},
)
def stat_files(files: list):
    """
    Check which files exist in the given list

    :param files: A list of file id to check
    :return: A list of files that do not exist.
    """


# Manage resources on the server


@method(
    path="/resource/<id>",
    operation="GET",
    agent_server=True,
    validate_sid=False,
    arg_options=ENV_OPTS,
    api=True,
    client_types=[const.ClientType.api, const.ClientType.agent],
)
def get_resource(
    tid: uuid.UUID, id: str, logs: bool = None, status: bool = None, log_action: const.ResourceAction = None, log_limit: int = 0
):
    """
    Return a resource with the given id.

    :param tid: The id of the environment this resource belongs to
    :param id: Get the resource with the given resource version id
    :param logs: Include the logs in the response
    :param status: Only return the status of the resource
    :param log_action: The log action to include, leave empty/none for all actions. Valid actions are one of
                      the action strings in const.ResourceAction
    :param log_limit: Limit the number of logs included in the response, up to a maximum of 1000.
                      To retrieve more entries, use  /api/v2/resource_actions
                      (:func:`~inmanta.protocol.methods_v2.get_resource_actions`)
    """


@method(path="/resource", operation="GET", agent_server=True, arg_options=ENV_OPTS, client_types=[const.ClientType.agent])
def get_resources_for_agent(
    tid: uuid.UUID, agent: str, sid: uuid.UUID = None, version: int = None, incremental_deploy: bool = False
):
    """
    Return the most recent state for the resources associated with agent, or the version requested

    :param tid: The id of the environment this resource belongs to
    :param agent: The agent
    :param sid: Session id of the agent (transparently added by agent client)
    :param version: The version to retrieve. If none, the latest available version is returned. With a specific version
                    that version is returned, even if it has not been released yet.
    :param incremental_deploy: Indicates whether the server should only return the resources that changed since the
                               previous deployment.
    """


@method(path="/resource", operation="POST", agent_server=True, arg_options=ENV_OPTS, client_types=[const.ClientType.agent])
def resource_action_update(
    tid: uuid.UUID,
    resource_ids: list,
    action_id: uuid.UUID,
    action: const.ResourceAction,
    started: datetime.datetime = None,
    finished: datetime.datetime = None,
    status: Optional[Union[const.ResourceState, const.DeprecatedResourceState]] = None,
    messages: list = [],
    changes: dict = {},
    change: const.Change = None,
    send_events: bool = False,
):
    """
    Send a resource update to the server

    :param tid: The id of the environment this resource belongs to
    :param resource_ids: The resource with the given resource_version_id id from the agent
    :param action_id: A unique id to indicate the resource action that has be updated
    :param action: The action performed
    :param started: The timestamp when this action was started. When this action (action_id) has not been saved yet,
                    started has to be defined.
    :param finished: The timestamp when this action was finished. Afterwards, no changes with the same action_id
                    can be stored. The status field also has to be set.
    :param status: The current status of the resource (if known)
    :param messages: A list of log entries to add to this entry.
    :param change:s A dict of changes to this resource. The key of this dict indicates the attributes/fields that
                   have been changed. The value contains the new value and/or the original value.
    :param change: The result of the changes
    :param send_events: [DEPRECATED] The value of this field is not used anymore.
    """


# Manage configuration model versions


@method(path="/version", operation="GET", arg_options=ENV_OPTS, client_types=[const.ClientType.api])
def list_versions(tid: uuid.UUID, start: int = None, limit: int = None):
    """
    Returns a list of all available versions

    :param tid: The id of the environment
    :param start: Optional, parameter to control the amount of results that are returned. 0 is the latest version.
    :param limit: Optional, parameter to control the amount of results returned, up to a maximum of 1000.
    """


@method(path="/version/<id>", operation="GET", arg_options=ENV_OPTS, client_types=[const.ClientType.api])
def get_version(tid: uuid.UUID, id: int, include_logs: bool = None, log_filter: str = None, limit: int = None):
    """
    Get a particular version and a list of all resources in this version

    :param tid: The id of the environment
    :param id: The id of the version to retrieve
    :param include_logs: If true, a log of all operations on all resources is included
    :param log_filter: Filter log to only include actions of the specified type
    :param limit: The maximal number of actions to return per resource (starting from the latest),
                    up to a maximum of 1000.
                    To retrieve more entries, use /api/v2/resource_actions
                    (:func:`~inmanta.protocol.methods_v2.get_resource_actions`)
    """


@method(path="/version/<id>", operation="DELETE", arg_options=ENV_OPTS, client_types=[const.ClientType.api])
def delete_version(tid: uuid.UUID, id: int):
    """
    Delete a particular version and resources

    :param tid: The id of the environment
    :param id: The id of the version to retrieve
    """


@method(path="/version", operation="PUT", arg_options=ENV_OPTS, client_types=[const.ClientType.compiler])
def put_version(
    tid: uuid.UUID,
    version: int,
    resources: list,
    resource_state: dict = {},
    unknowns: list = None,
    version_info: dict = None,
    compiler_version: str = None,
    resource_sets: Dict[model.ResourceIdStr, Optional[str]] = {},
):
    """
    Store a new version of the configuration model

    The version number must be obtained through the reserve_version call

    :param tid: The id of the environment
    :param version: The version of the configuration model
    :param resources: A list of all resources in the configuration model (deployable)
    :param resource_state: A dictionary with the initial const.ResourceState per resource id. The ResourceState should be set
                           to undefined when the resource depends on an unknown or available when it doesn't.
    :param unknowns: A list of unknown parameters that caused the model to be incomplete
    :param version_info: Module version information
    :param compiler_version: version of the compiler, if not provided, this call will return an error
    :param resource_sets: a dictionary describing which resource belongs to which resource set
    """


@method(
    path="/version/<id>", operation="POST", arg_options=ENV_OPTS, client_types=[const.ClientType.api, const.ClientType.compiler]
)
def release_version(tid: uuid.UUID, id: int, push: bool = False, agent_trigger_method: const.AgentTriggerMethod = None):
    """
    Release version of the configuration model for deployment.

    :param tid: The id of the environment
    :param id: The version of the CM to deploy
    :param push: Notify all agents to deploy the version
    :param agent_trigger_method: Indicates whether the agents should perform a full or an incremental deploy when
                                push is true.
    """


@method(path="/deploy", operation="POST", arg_options=ENV_OPTS, client_types=[const.ClientType.api])
def deploy(
    tid: uuid.UUID,
    agent_trigger_method: const.AgentTriggerMethod = const.AgentTriggerMethod.push_full_deploy,
    agents: list = None,
):
    """
    Notify agents to perform a deploy now.

    :param tid: The id of the environment.
    :param agent_trigger_method: Indicates whether the agents should perform a full or an incremental deploy.
    :param agents: Optional, names of specific agents to trigger
    """


# Method for requesting and quering a dryrun


@method(path="/dryrun/<id>", operation="POST", arg_options=ENV_OPTS, client_types=[const.ClientType.api])
def dryrun_request(tid: uuid.UUID, id: int):
    """
    Do a dryrun

    :param tid: The id of the environment
    :param id: The version of the CM to deploy
    """


@method(path="/dryrun", operation="GET", arg_options=ENV_OPTS, client_types=[const.ClientType.api])
def dryrun_list(tid: uuid.UUID, version: int = None):
    """
    Get the list of dry runs for an environment. The results are sorted by dry run id.

    :param tid: The id of the environment
    :param version: Only for this version
    """


@method(path="/dryrun/<id>", operation="GET", arg_options=ENV_OPTS, client_types=[const.ClientType.api])
def dryrun_report(tid: uuid.UUID, id: uuid.UUID):
    """
    Create a dryrun report

    :param tid: The id of the environment
    :param id: The version dryrun to report
    """


@method(path="/dryrun/<id>", operation="PUT", agent_server=True, arg_options=ENV_OPTS, client_types=[const.ClientType.agent])
def dryrun_update(tid: uuid.UUID, id: uuid.UUID, resource: str, changes: dict):
    """
    Store dryrun results at the server

    :param tid: The id of the environment
    :param id: The version dryrun to report
    :param resource: The id of the resource
    :param changes: The required changes
    """


# Method for requesting a dryrun from an agent


@method(path="/agent_dryrun/<id>", operation="POST", server_agent=True, timeout=5, arg_options=AGENT_ENV_OPTS, client_types=[])
def do_dryrun(tid: uuid.UUID, id: uuid.UUID, agent: str, version: int):
    """
    Do a dryrun on an agent

    :param tid: The environment id
    :param id: The id of the dryrun
    :param agent: The agent to do the dryrun for
    :param version: The version of the model to dryrun
    """


# Method to notify the server of changes in the configuration model source code


@method(
    path="/notify/<id>",
    operation="GET",
    arg_options={"id": ArgOption(getter=convert_environment)},
    client_types=[const.ClientType.api],
)
def notify_change_get(id: uuid.UUID, update: bool = True):
    """
    Simplified GET version of the POST method
    """


@method(
    path="/notify/<id>",
    operation="POST",
    arg_options={"id": ArgOption(getter=convert_environment)},
    client_types=[const.ClientType.api],
)
def notify_change(id: uuid.UUID, update: bool = True, metadata: dict = {}):
    """
    Notify the server that the repository of the environment with the given id, has changed.

    :param id: The id of the environment
    :param update: Update the model code and modules. Default value is true
    :param metadata: The metadata that indicates the source of the compilation trigger.
    """


@method(path="/notify/<id>", operation="HEAD", client_types=[const.ClientType.api])
def is_compiling(id: uuid.UUID):
    """
    Is a compiler running for the given environment

    :param id: The environment id
    """


# Get and set parameters on the server


@method(
    path="/parameter/<id>",
    operation="GET",
    arg_options=ENV_OPTS,
    client_types=[const.ClientType.api, const.ClientType.compiler, const.ClientType.agent],
)
def get_param(tid: uuid.UUID, id: str, resource_id: str = None):
    """
    Get a parameter from the server.

    :param tid: The id of the environment
    :param id: The name of the parameter
    :param resource_id: Optionally, scope the parameter to resource (fact),
                        if the resource id should not contain a version, the latest version is used
    :return: Returns the following status codes:
            200: The parameter content is returned
            404: The parameter is not found and unable to find it because its resource is not known to the server
            410: The parameter has expired
            503: The parameter is not found but its value is requested from an agent
    """


@method(
    path="/parameter/<id>",
    operation="PUT",
    arg_options=ENV_OPTS,
    client_types=[const.ClientType.api, const.ClientType.compiler, const.ClientType.agent],
)
def set_param(
    tid: uuid.UUID,
    id: str,
    source: const.ParameterSource,
    value: str,
    resource_id: str = None,
    metadata: dict = {},
    recompile: bool = False,
):
    """
    Set a parameter on the server. If the parameter is an tracked unknown, it will trigger a recompile on the server.
    Otherwise, if the value is changed and recompile is true, a recompile is also triggered.

    :param tid: The id of the environment
    :param id: The name of the parameter
    :param resource_id: Optionally, scope the parameter to resource (fact)
    :param source: The source of the parameter.
    :param value: The value of the parameter
    :param metadata: metadata about the parameter
    :param recompile: Whether to trigger a recompile
    """


@method(
    path="/parameter/<id>",
    operation="DELETE",
    arg_options=ENV_OPTS,
    client_types=[const.ClientType.api, const.ClientType.compiler, const.ClientType.agent],
)
def delete_param(tid: uuid.UUID, id: str, resource_id: str = None):
    """
    Delete a parameter on the server

    :param tid: The id of the environment
    :param id: The name of the parameter
    :param resource_id: The resource id of the parameter
    """


@method(
    path="/parameter", operation="POST", arg_options=ENV_OPTS, client_types=[const.ClientType.api, const.ClientType.compiler]
)
def list_params(tid: uuid.UUID, query: dict = {}):
    """
    List/query parameters in this environment. The results are ordered alphabetically by parameter name.

    :param tid: The id of the environment
    :param query: A query to match against metadata
    """


#  Get and set parameters on the server


@method(
    path="/parameters",
    operation="PUT",
    agent_server=True,
    arg_options=ENV_OPTS,
    client_types=[const.ClientType.api, const.ClientType.compiler, const.ClientType.agent],
)
def set_parameters(tid: uuid.UUID, parameters: list):
    """
    Set a parameter on the server

    :param tid: The id of the environment
    :param parameters: A list of dicts with the following keys:
        - id The name of the parameter
        - source The source of the parameter. Valid values are defined in the ParameterSource enum (see: inmanta/const.py)
        - value The value of the parameter
        - resource_id Optionally, scope the parameter to resource (fact)
        - metadata metadata about the parameter
    """


# Get parameters from the agent


@method(path="/agent_parameter", operation="POST", server_agent=True, timeout=5, arg_options=AGENT_ENV_OPTS, client_types=[])
def get_parameter(tid: uuid.UUID, agent: str, resource: dict):
    """
    Get all parameters/facts known by the agents for the given resource

    :param tid: The environment
    :param agent: The agent to get the parameters from
    :param resource: The resource to query the parameters from
    """


@method(path="/code/<id>", operation="GET", agent_server=True, arg_options=ENV_OPTS, client_types=[const.ClientType.agent])
def get_code(tid: uuid.UUID, id: int, resource: str):
    """
    Get the code for a given version of the configuration model

    :param tid: The environment the code belongs to
    :param id: The id (version) of the configuration model
    """


@method(path="/codebatched/<id>", operation="PUT", arg_options=ENV_OPTS, client_types=[const.ClientType.compiler])
def upload_code_batched(tid: uuid.UUID, id: int, resources: dict):
    """
    Upload the supporting code to the server

    :param tid: The environment the code belongs to
    :param id: The id (version) of the configuration model
    :param resource: a dict mapping resources to dicts mapping file names to file hashes
    """


# Generate download the diff of two hashes


@method(path="/filediff", client_types=[const.ClientType.api])
def diff(a: str, b: str):
    """
    Returns the diff of the files with the two given ids
    """


# Get a list of compile reports


@method(path="/compilereport", operation="GET", arg_options=ENV_OPTS, client_types=[const.ClientType.api])
def get_reports(tid: uuid.UUID, start: str = None, end: str = None, limit: int = None):
    """
    Return compile reports newer then start

    :param tid: The id of the environment to get a report from
    :param start: Reports after start
    :param end: Reports before end
    :param limit: Maximum number of results, up to a maximum of 1000
    """


@method(path="/compilereport/<id>", operation="GET", client_types=[const.ClientType.api])
def get_report(id: uuid.UUID):
    """
    Get a compile report from the server

    :param id: The id of the compile and its reports to fetch.
    """


# Get a list of all agents


@method(path="/agentproc", operation="GET", client_types=[const.ClientType.api])
def list_agent_processes(
    environment: uuid.UUID = None, expired: bool = True, start: uuid.UUID = None, end: uuid.UUID = None, limit: int = None
):
    """
    Return a list of all nodes and the agents for these nodes

    :param environment: An optional environment. If set, only the agents that belong to this environment are returned
    :param expired: Optional, also show expired processes, otherwise only living processes are shown.
    :param start: Agent processes after start (sorted by sid in ASC)
    :param end: Agent processes before end (sorted by sid in ASC)
    :param limit: Maximum number of results, up to a maximum of 1000

    :raises BadRequest: limit parameter can not exceed 1000
    :raises NotFound: The given environment id does not exist!

    :return: A list of nodes
    """


@method(path="/agentproc/<id>", operation="GET", client_types=[const.ClientType.api])
def get_agent_process(id: uuid.UUID):
    """
    Return a detailed report for a node

    :param id: The session id of the agent
    :return: The requested node
    """


# Get a list of all agents
@method(path="/agent/<id>", operation="POST", api=True, timeout=5, arg_options=ENV_OPTS, client_types=[const.ClientType.api])
def trigger_agent(tid: uuid.UUID, id: str):
    """
    Request the server to reload an agent

    :param tid: The environment this agent is defined in
    :param id: The name of the agent
    :return: The requested node
    """


@method(path="/agent", operation="GET", api=True, timeout=5, arg_options=ENV_OPTS, client_types=[const.ClientType.api])
def list_agents(tid: uuid.UUID, start: str = None, end: str = None, limit: int = None):
    """
    List all agent for an environment

    :param tid: The environment the agents are defined in
    :param start: Agent after start (sorted by name in ASC)
    :param end: Agent before end (sorted by name in ASC)
    :param limit: Maximum number of results, up to a maximum of 1000

    :raises BadRequest: limit parameter can not exceed 1000
    :raises NotFound: The given environment id does not exist!
    """


# Reporting by the agent to the server


@method(path="/status", operation="GET", server_agent=True, timeout=5, client_types=[])
def get_status():
    """
    A call from the server to the agent to report its status to the server

    :return: A map with report items
    """


# Methods to allow the server to set the agents state


@method(path="/agentstate", operation="POST", server_agent=True, timeout=5, client_types=[])
def set_state(agent: str, enabled: bool):
    """
    Set the state of the agent.
    """


@method(path="/agentstate/<id>", operation="POST", server_agent=True, timeout=5, arg_options=AGENT_ENV_OPTS, client_types=[])
def trigger(tid: uuid.UUID, id: str, incremental_deploy: bool):
    """
    Request an agent to reload resources

    :param tid: The environment this agent is defined in
    :param id: The name of the agent
    :param incremental_deploy: Indicates whether the agent should perform an incremental deploy or a full deploy
    """


# Methods to send event to the server


@method(path="/event/<id>", operation="PUT", server_agent=True, timeout=5, arg_options=AGENT_ENV_OPTS, client_types=[])
def resource_event(
    tid: uuid.UUID, id: str, resource: str, send_events: bool, state: const.ResourceState, change: const.Change, changes={}
):
    """
    Tell an agent a resource it waits for has been updated

    :param tid: The environment this agent is defined in
    :param id: The name of the agent
    :param resource: The resource ID of the resource being updated
    :param send_events: [DEPRECATED] The value of this field is not used anymore.
    :param state: State the resource acquired (deployed, skipped, canceled)
    :param change: The change that was made to the resource
    :param changes: The changes made to the resource
    """


# Methods for the agent to get its initial state from the server


@method(path="/agentrecovery", operation="GET", agent_server=True, arg_options=ENV_OPTS, client_types=[const.ClientType.agent])
def get_state(tid: uuid.UUID, sid: uuid.UUID, agent: str):
    """
    Get the state for this agent.

    :return: A map with key enabled and value a boolean.
    """


@typedmethod(path="/serverstatus", operation="GET", client_types=[const.ClientType.api])
def get_server_status() -> model.StatusResponse:
    """
    Get the status of the server
    """


@typedmethod(
    path="/compilequeue",
    operation="GET",
    arg_options=ENV_OPTS,
    client_types=[const.ClientType.api],
    api_version=1,
    envelope_key="queue",
)
def get_compile_queue(tid: uuid.UUID) -> List[model.CompileRun]:
    """
    Get the current compiler queue on the server, ordered by increasing `requested` timestamp.
    """



"""
    Copyright 2021 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: bart@inmanta.com
"""
import glob
import importlib
import json
import logging
import os
import re
import subprocess
import sys
import tempfile
from importlib.abc import Loader
from subprocess import CalledProcessError
from typing import Dict, List, Optional, Pattern, Tuple
from unittest.mock import patch

import py
import pytest
from pkg_resources import Requirement

from inmanta import env, loader, module
from packaging import version
from utils import LogSequence, PipIndex, create_python_package

if "inmanta-core" in env.process_env.get_installed_packages(only_editable=True):
    pytest.skip(
        "The tests in this module will fail if it runs against inmanta-core installed in editable mode, "
        "because the build tag on the development branch is set to .dev0 by default. The inmanta package protection feature "
        "would make pip install a non-editable version of the same package. But no version with build tag .dev0 exists "
        "on the python package repository.",
        allow_module_level=True,
    )


@pytest.mark.slowtest
def test_venv_pyton_env_empty_string(tmpdir):
    """test that an exception is raised if the venv path is an empty string"""
    with pytest.raises(ValueError) as e:
        env.VirtualEnv("")
    assert e.value.args[0] == "The env_path cannot be an empty string."

    env_dir1 = tmpdir.mkdir("env1").strpath
    venv1 = env.VirtualEnv(env_dir1)
    venv1.use_virtual_env()

    env_dir2 = tmpdir.mkdir("env2").strpath
    venv2 = env.VirtualEnv(env_dir2)
    venv2.env_path = ""
    with pytest.raises(Exception) as e:
        venv2.use_virtual_env()
    assert e.value.args[0] == "The env_path cannot be an empty string."

    with pytest.raises(ValueError) as e:
        env.PythonEnvironment(python_path="")
    assert e.value.args[0] == "The python_path cannot be an empty string."

    with pytest.raises(ValueError) as e:
        env.PythonEnvironment(env_path="")
    assert e.value.args[0] == "The env_path cannot be an empty string."


@pytest.mark.slowtest
def test_basic_install(tmpdir):
    """If this test fails, try running "pip uninstall lorem dummy-yummy iplib" before running it."""
    env_dir1 = tmpdir.mkdir("env1").strpath

    with pytest.raises(ImportError):
        import lorem  # NOQA

    venv1 = env.VirtualEnv(env_dir1)

    venv1.use_virtual_env()
    venv1.install_from_list(["lorem"])
    import lorem  # NOQA

    lorem.sentence()

    with pytest.raises(ImportError):
        import yummy  # NOQA

    venv1 = env.VirtualEnv(env_dir1)

    venv1.use_virtual_env()
    venv1.install_from_list(["dummy-yummy"])
    import yummy  # NOQA

    with pytest.raises(ImportError):
        import iplib  # NOQA

    venv1 = env.VirtualEnv(env_dir1)

    venv1.use_virtual_env()
    try:
        venv1.install_from_list(
            ["lorem == 0.1.1", "dummy-yummy", "iplib@git+https://github.com/bartv/python3-iplib", "lorem", "iplib >=0.0.1"]
        )
    except CalledProcessError as ep:
        print(ep.stdout)
        raise
    import iplib  # NOQA


@pytest.mark.slowtest
def test_install_fails(tmpdir, caplog, monkeypatch):
    venv = env.VirtualEnv(tmpdir)
    venv.use_virtual_env()
    caplog.clear()
    caplog.set_level(logging.INFO)
    package_name = "non-existing-pkg-inmanta"

    # monkeypatch pip install to set --no-index for security reasons (anyone could publish this package to PyPi)
    compose = env.PipCommandBuilder.compose_install_command

    def mock_compose(*args, **kwargs):
        if "index_urls" in kwargs or len(args) < 5:
            return compose(*args, **{**kwargs, "index_urls": []})
        else:
            return compose(*args[:3], [], *args[4:], **kwargs)

    monkeypatch.setattr(env.PipCommandBuilder, "compose_install_command", mock_compose)

    with pytest.raises(Exception):
        venv.install_from_list([package_name])

    log_sequence = LogSequence(caplog)
    log_sequence.contains("inmanta.env", logging.INFO, f"requirements:\n{package_name}")


@pytest.mark.slowtest
def test_install_package_already_installed_in_parent_env(tmpdir):
    """Test using and installing a package that is already present in the parent virtual environment."""
    # get all packages in the parent
    parent_installed = list(env.process_env.get_installed_packages().keys())

    # create a venv and list all packages available in the venv
    venv = env.VirtualEnv(tmpdir)
    venv.use_virtual_env()

    installed_packages = list(env.PythonEnvironment(python_path=venv._parent_python).get_installed_packages().keys())

    # verify that the venv sees all parent packages
    assert not set(parent_installed) - set(installed_packages)

    # site dir should be empty
    site_dir = os.path.join(venv.env_path, "lib/python*/site-packages")
    dirs = glob.glob(site_dir)
    assert len(dirs) == 1
    site_dir = dirs[0]

    def _list_dir(path: str, ignore: List[str]) -> List[str]:
        return [d for d in os.listdir(site_dir) if d not in ignore]

    # site_dir should only contain a sitecustomize.py file that sets up inheritance from the parent venv
    assert not _list_dir(site_dir, ignore=["inmanta-inherit-from-parent-venv.pth", "__pycache__"])

    # test installing a package that is already present in the parent venv
    assert "more-itertools" in parent_installed
    venv.install_from_list(["more-itertools"])

    # site_dir should only contain a sitecustomize.py file that sets up inheritance from the parent venv
    assert not _list_dir(site_dir, ignore=["inmanta-inherit-from-parent-venv.pth", "__pycache__"])

    # report json
    subprocess.check_output([os.path.join(venv.env_path, "bin/pip"), "list"])


def test_req_parser(tmpdir):
    url = "git+https://github.com/bartv/python3-iplib"
    at_url = "iplib@" + url
    egg_url = url + "#egg=iplib"

    e = env.VirtualEnv(tmpdir)
    name, u = e._parse_line(url)
    assert name is None
    assert u == url

    name, u = e._parse_line(at_url)
    assert name == "iplib"
    assert u == egg_url

    e._parse_line(egg_url)
    assert name == "iplib"
    assert u == egg_url


def test_gen_req_file(tmpdir):
    e = env.VirtualEnv(tmpdir)
    req = [
        "lorem == 0.1.1",
        "lorem > 0.1",
        "dummy-yummy",
        "iplib@git+https://github.com/bartv/python3-iplib",
        "lorem",
        # verify support for environment markers as described in PEP 508
        "lorem;python_version<'3.7'",
        "lorem;platform_machine == 'x86_64' and platform_system == 'Linux'",
    ]

    req_lines = [x for x in e._gen_content_requirements_file(req).split("\n") if len(x) > 0]
    assert len(req_lines) == 3
    assert (
        'lorem == 0.1.1, > 0.1 ; python_version < "3.7" and platform_machine == "x86_64" and platform_system == "Linux"'
        in req_lines
    )


def test_environment_python_version_multi_digit(tmpdir: py.path.local) -> None:
    """
    Make sure the constructor for env.Environment can handle multi-digit minor versions of Python to ensure compatibility with
    Python 3.10+.
    """
    with patch("sys.version_info", new=(3, 123, 0)):
        # python version is not included in path on windows
        with patch("sys.platform", new="linux"):
            assert env.PythonEnvironment(env_path=str(tmpdir)).site_packages_dir == os.path.join(
                str(tmpdir), "lib", "python3.123", "site-packages"
            )


@pytest.mark.slowtest
@pytest.mark.parametrize_any("version", [None, version.Version("8.6.0")])
def test_process_env_install_from_index(
    tmpdir: str,
    tmpvenv_active: Tuple[py.path.local, py.path.local],
    version: Optional[version.Version],
) -> None:
    """
    Install a package from a pip index into the process_env. Assert any version specs are respected.
    """
    package_name: str = "more-itertools"
    assert package_name not in env.process_env.get_installed_packages()
    env.process_env.install_from_index([Requirement.parse(package_name + (f"=={version}" if version is not None else ""))])
    installed: Dict[str, version.Version] = env.process_env.get_installed_packages()
    assert package_name in installed
    if version is not None:
        assert installed[package_name] == version


@pytest.mark.slowtest
def test_process_env_install_from_index_not_found(tmpvenv_active: Tuple[py.path.local, py.path.local]) -> None:
    """
    Attempt to install a package that does not exist from a pip index. Assert the appropriate error is raised.
    """
    with pytest.raises(env.PackageNotFound):
        # pass empty index list for security reasons (anyone could publish this package to PyPi)
        env.process_env.install_from_index([Requirement.parse("this-package-does-not-exist")], index_urls=[])


@pytest.mark.slowtest
def test_process_env_install_from_index_conflicting_reqs(
    tmpdir: str, tmpvenv_active: Tuple[py.path.local, py.path.local]
) -> None:
    """
    Attempt to install a package with conflicting version requirements from a pip index. Make sure this fails and the
    package remains uninstalled.
    """
    package_name: str = "more-itertools"
    with pytest.raises(env.ConflictingRequirements) as e:
        env.process_env.install_from_index([Requirement.parse(f"{package_name}{version}") for version in [">8.5", "<=8"]])
    assert "conflicting dependencies" in e.value.msg
    assert package_name not in env.process_env.get_installed_packages()


@pytest.mark.slowtest
@pytest.mark.parametrize("editable", [True, False])
def test_process_env_install_from_source(
    tmpvenv_active: Tuple[py.path.local, py.path.local],
    modules_v2_dir: str,
    editable: bool,
) -> None:
    """
    Install a package from source into the process_env. Make sure the editable option actually results in an editable install.
    """
    package_name: str = "inmanta-module-minimalv2module"
    project_dir: str = os.path.join(modules_v2_dir, "minimalv2module")
    assert package_name not in env.process_env.get_installed_packages()
    env.process_env.install_from_source([env.LocalPackagePath(path=project_dir, editable=editable)])
    assert package_name in env.process_env.get_installed_packages()
    if editable:
        assert package_name in env.process_env.get_installed_packages(only_editable=True)


# v1 plugin loader overrides loader paths so verify that it doesn't interfere with env.process_env installs
@pytest.mark.parametrize("v1_plugin_loader", [True, False])
@pytest.mark.parametrize("package_name", ["lorem", "more-itertools", "inmanta-module-minimalv2module"])
@pytest.mark.slowtest
def test_active_env_get_module_file(
    local_module_package_index: str,
    tmpdir: py.path.local,
    tmpvenv_active: Tuple[py.path.local, py.path.local],
    v1_plugin_loader: bool,
    package_name: str,
) -> None:
    """
    Test the env.ActiveEnv.get_module_file() command on a newly installed package. Make sure it works regardless of whether we
    install a dependency of inmanta-core (which is already installed in the encapsulating development venv), a new package or an
    inmanta module (namespace package).
    """
    venv_dir, _ = tmpvenv_active

    if package_name.startswith(module.ModuleV2.PKG_NAME_PREFIX):
        module_name = "inmanta_plugins." + package_name[len(module.ModuleV2.PKG_NAME_PREFIX) :].replace("-", "_")
        index = str(local_module_package_index)
    else:
        module_name = package_name.replace("-", "_")
        index = None

    # unload module if already loaded from encapsulating development venv
    if module_name in sys.modules:
        loaded = [sub for sub in sys.modules.keys() if sub.startswith(module_name)]
        for sub in loaded:
            del sys.modules[sub]
    importlib.invalidate_caches()

    if v1_plugin_loader:
        loader.PluginModuleFinder.configure_module_finder([os.path.join(str(tmpdir), "libs")])

    assert env.ActiveEnv.get_module_file(module_name) is None
    env.process_env.install_from_index([Requirement.parse(package_name)], index_urls=[index] if index is not None else None)
    assert package_name in env.process_env.get_installed_packages()
    module_info: Optional[Tuple[Optional[str], Loader]] = env.ActiveEnv.get_module_file(module_name)
    assert module_info is not None
    module_file, mod_loader = module_info
    assert module_file is not None
    assert not isinstance(mod_loader, loader.PluginModuleLoader)
    assert module_file == os.path.join(env.process_env.site_packages_dir, *module_name.split("."), "__init__.py")
    # verify that the package was installed in the development venv
    assert str(venv_dir) in module_file
    importlib.import_module(module_name)
    assert module_name in sys.modules
    assert sys.modules[module_name].__file__ == module_file


@pytest.mark.slowtest
def test_active_env_get_module_file_editable_namespace_package(
    tmpdir: str,
    tmpvenv_active: Tuple[py.path.local, py.path.local],
    modules_v2_dir: str,
) -> None:
    """
    Verify that get_module_file works after installing an editable namespace package in an active environment.
    """
    package_name: str = "inmanta-module-minimalv2module"
    module_name: str = "inmanta_plugins.minimalv2module"

    assert env.ActiveEnv.get_module_file(module_name) is None
    project_dir: str = os.path.join(modules_v2_dir, "minimalv2module")
    env.process_env.install_from_source([env.LocalPackagePath(path=project_dir, editable=True)])
    assert package_name in env.process_env.get_installed_packages()
    module_info: Optional[Tuple[Optional[str], Loader]] = env.ActiveEnv.get_module_file(module_name)
    assert module_info is not None
    module_file, mod_loader = module_info
    assert module_file is not None
    assert not isinstance(mod_loader, loader.PluginModuleLoader)
    assert module_file == os.path.join(modules_v2_dir, "minimalv2module", *module_name.split("."), "__init__.py")
    importlib.import_module(module_name)
    assert module_name in sys.modules
    assert sys.modules[module_name].__file__ == module_file


def create_install_package(name: str, version: version.Version, requirements: List[Requirement]) -> None:
    """
    Creates and installs a simple package with specified requirements. Creates package in a temporary directory and
    cleans it up after install.

    :param name: Package name.
    :param version: Version for this package.
    :param requirements: Requirements on other packages. Required packages must already be installed when calling this function.
    """
    req_string: str = (
        "" if len(requirements) == 0 else ("[options]\ninstall_requires=" + "\n    ".join(str(req) for req in requirements))
    )
    with tempfile.TemporaryDirectory() as tmpdir:
        with open(os.path.join(tmpdir, "setup.cfg"), "w") as fd:
            fd.write(
                f"""
[metadata]
name = {name}
version = {version}

{req_string}
                """.strip()
            )
        with open(os.path.join(tmpdir, "pyproject.toml"), "w") as fd:
            fd.write(
                """
[build-system]
requires = ["setuptools", "wheel"]
build-backend = "setuptools.build_meta"
                """.strip()
            )
        env.process_env.install_from_source([env.LocalPackagePath(path=str(tmpdir), editable=False)])


@pytest.mark.slowtest
def test_active_env_check_basic(
    caplog,
    tmpdir: str,
    tmpvenv_active_inherit: str,
) -> None:
    """
    Verify that the env.ActiveEnv.check() method detects all possible forms of incompatibilities within the environment.
    """
    caplog.set_level(logging.WARNING)

    in_scope_test: Pattern[str] = re.compile("test-package-.*")
    in_scope_nonext: Pattern[str] = re.compile("nonexistant-package")

    error_msg: str = "Incompatibility between constraint"

    def assert_all_checks(expect_test: Tuple[bool, str] = (True, ""), expect_nonext: Tuple[bool, str] = (True, "")) -> None:
        """
        verify what the check method for 2 different scopes: for an existing package and a non existing one.

        param: expect_test: Tuple with as first value a bool and as second value a string. The bool is true if the execution
        will not raise an error, false if it will raise an error. The second argument is the warning message that can be find
        in the logs.
        param: expect_nonext: Tuple with as first value a bool and as second value a string. The bool is true if the execution
        will not raise an error, false if it will raise an error. The second argument is the warning message that can be find
        in the logs.
        """
        for in_scope, expect in [(in_scope_test, expect_test), (in_scope_nonext, expect_nonext)]:
            caplog.clear()
            if expect[0]:
                env.ActiveEnv.check(in_scope)
                if expect[1] == "":
                    assert error_msg not in {rec.message for rec in caplog.records}
                else:
                    assert expect[1] in {rec.message for rec in caplog.records}
            else:
                with pytest.raises(env.ConflictingRequirements) as e:
                    env.ActiveEnv.check(in_scope)
                assert expect[1] in e.value.get_message()

    assert_all_checks()
    create_install_package("test-package-one", version.Version("1.0.0"), [])
    assert_all_checks()
    create_install_package("test-package-two", version.Version("1.0.0"), [Requirement.parse("test-package-one~=1.0")])
    assert_all_checks()
    create_install_package("test-package-one", version.Version("2.0.0"), [])
    assert_all_checks(
        expect_test=(
            False,
            "Incompatibility between constraint test-package-one~=1.0 and installed version 2.0.0 (from test-package-two)",
        ),
        expect_nonext=(True, error_msg + " test-package-one~=1.0 and installed version 2.0.0 (from test-package-two)"),
    )


@pytest.mark.slowtest
def test_active_env_check_constraints(caplog, tmpvenv_active_inherit: str) -> None:
    """
    Verify that the env.ActiveEnv.check() method's constraints parameter is taken into account as expected.
    """
    caplog.set_level(logging.WARNING)
    in_scope: Pattern[str] = re.compile("test-package-.*")
    constraints: List[Requirement] = [Requirement.parse("test-package-one~=1.0")]

    env.ActiveEnv.check(in_scope)

    caplog.clear()
    with pytest.raises(env.ConflictingRequirements):
        env.ActiveEnv.check(in_scope, constraints)

    caplog.clear()
    create_install_package("test-package-one", version.Version("1.0.0"), [])
    env.ActiveEnv.check(in_scope, constraints)
    assert "Incompatibility between constraint" not in caplog.text

    # Add an unrelated package to the venv, that should not matter
    # setup for #4761
    caplog.clear()
    create_install_package("ext-package-one", version.Version("1.0.0"), [Requirement.parse("test-package-one==1.0")])
    env.ActiveEnv.check(in_scope, constraints)
    assert "Incompatibility between constraint" not in caplog.text

    caplog.clear()
    v: version.Version = version.Version("2.0.0")
    create_install_package("test-package-one", v, [])
    # test for #4761
    # without additional constrain, this is not a hard failure
    # except for the unrelated package, which should produce a warning
    env.ActiveEnv.check(in_scope, [])
    assert (
        "Incompatibility between constraint test-package-one==1.0 and installed version 2.0.0 (from ext-package-one)"
        in caplog.text
    )

    caplog.clear()
    with pytest.raises(env.ConflictingRequirements):
        env.ActiveEnv.check(in_scope, constraints)


@pytest.mark.slowtest
def test_override_inmanta_package(tmpvenv_active_inherit: env.VirtualEnv) -> None:
    """
    Ensure that an ActiveEnv cannot override the main inmanta packages: inmanta-service-orchestrator, inmanta, inmanta-core.
    """
    installed_pkgs = tmpvenv_active_inherit.get_installed_packages()
    assert "inmanta-core" in installed_pkgs, "The inmanta-core package should be installed to run the tests"

    inmanta_requirements = Requirement.parse("inmanta-core==4.0.0")
    with pytest.raises(env.ConflictingRequirements) as excinfo:
        tmpvenv_active_inherit.install_from_index(requirements=[inmanta_requirements])
    match = re.search(
        r"Cannot install inmanta-core==4\.0\.0 and inmanta-core=.* because these "
        r"package versions have conflicting dependencies",
        excinfo.value.msg,
    )
    assert match is not None


@pytest.mark.slowtest
def test_pip_binary_when_venv_path_contains_double_quote(tmpdir) -> None:
    """
    Test whether the pip binary generated by the VirtualEnv class works correctly when the
    pip binary contains a double quote.
    """
    venv_dir = os.path.join(tmpdir, 'tes t"test')
    venv = env.VirtualEnv(venv_dir)
    venv.use_virtual_env()
    assert any('"' in path and " " in path for path in sys.path)

    pip_binary = os.path.join(os.path.dirname(venv.python_path), "pip")
    # Ensure that the pip command doesn't raise an exception
    result = subprocess.check_output(
        [pip_binary, "list", "--format", "json", "--disable-pip-version-check", "--no-python-version-warning"],
        timeout=10,
        encoding="utf-8",
    )
    parsed_output = json.loads(result)
    # Ensure inheritance works correctly
    assert "inmanta-core" in [elem["name"] for elem in parsed_output]


@pytest.mark.slowtest
def test_cache_on_active_env(tmpvenv_active_inherit: env.ActiveEnv, local_module_package_index: str) -> None:
    """
    Test whether the cache on an active env works correctly.
    """

    def _assert_install(requirement: str, installed: bool) -> None:
        parsed_requirement = Requirement.parse(requirement)
        for r in [requirement, parsed_requirement]:
            assert tmpvenv_active_inherit.are_installed(requirements=[r]) == installed

    _assert_install("inmanta-module-elaboratev2module==1.2.3", installed=False)
    tmpvenv_active_inherit.install_from_index(
        requirements=[Requirement.parse("inmanta-module-elaboratev2module==1.2.3")],
        index_urls=[local_module_package_index],
    )
    _assert_install("inmanta-module-elaboratev2module==1.2.3", installed=True)
    _assert_install("inmanta-module-elaboratev2module~=1.2.0", installed=True)
    _assert_install("inmanta-module-elaboratev2module<1.2.4", installed=True)
    _assert_install("inmanta-module-elaboratev2module>1.2.3", installed=False)
    _assert_install("inmanta-module-elaboratev2module==1.2.4", installed=False)


def test_gen_content_requirements_file_extras():
    """
    Ensure that the `env.ActiveEnv._gen_content_requirements_file` method takes into account extras.
    """
    dependency = "dep==1.2.3"
    content: str = env.ActiveEnv._gen_content_requirements_file([dependency])
    assert content.strip() == "dep == 1.2.3"

    dependency = "dep[opt]==1.2.3"
    content: str = env.ActiveEnv._gen_content_requirements_file([dependency])
    assert content.strip() == "dep[opt] == 1.2.3"

    dependency = "dep[opt,dev]==1.2.3"
    content: str = env.ActiveEnv._gen_content_requirements_file([dependency])
    assert content.strip() == "dep[dev,opt] == 1.2.3"


@pytest.mark.slowtest
def test_basic_logging(tmpdir, caplog):
    with caplog.at_level(logging.INFO):
        env_dir1 = tmpdir.mkdir("env1").strpath

        venv1 = env.VirtualEnv(env_dir1)

        venv1.use_virtual_env()

        log_sequence = LogSequence(caplog)
        log_sequence.assert_not("inmanta.env", logging.INFO, f"Creating new virtual environment in {env_dir1}")
        log_sequence.contains("inmanta.env", logging.INFO, f"Using virtual environment at {env_dir1}")


@pytest.mark.slowtest
def test_are_installed_dependency_cycle_on_extra(tmpdir, tmpvenv_active_inherit: env.VirtualEnv) -> None:
    """
    Ensure that the `ActiveEnv.are_installed()` method doesn't go into an infinite loop when there is a circular dependency
    involving an extra.

    Dependency loop:
        pkg[optional]
           -> dep[optional]
               -> pkg[optional]
    """
    pip_index = PipIndex(artifact_dir=str(tmpdir))
    create_python_package(
        name="pkg",
        pkg_version=version.Version("1.0.0"),
        path=os.path.join(tmpdir, "pkg"),
        publish_index=pip_index,
        optional_dependencies={
            "optional-pkg": [Requirement.parse("dep[optional-dep]")],
        },
    )
    create_python_package(
        name="dep",
        pkg_version=version.Version("1.0.0"),
        path=os.path.join(tmpdir, "dep"),
        publish_index=pip_index,
        optional_dependencies={
            "optional-dep": [Requirement.parse("pkg[optional-pkg]")],
        },
    )

    requirements = [Requirement.parse("pkg[optional-pkg]")]
    tmpvenv_active_inherit.install_from_index(requirements=requirements, index_urls=[pip_index.url])
    assert tmpvenv_active_inherit.are_installed(requirements=requirements)


@pytest.mark.slowtest
def test_pip_logs(caplog, tmpvenv_active_inherit: str) -> None:
    """
    Verify the logs of a pip install:
        - all records start with 'inmanta.pip'
        - content of requirements and constraints files are logged
        - the pip command is logged
    """
    caplog.set_level(logging.DEBUG)

    with tempfile.TemporaryDirectory() as tmpdir:
        requirement1 = os.path.join(tmpdir, "requirement1.txt")
        requirement2 = os.path.join(tmpdir, "requirement2.txt")
        constraint1 = os.path.join(tmpdir, "constraint1.txt")
        constraint2 = os.path.join(tmpdir, "constraint2.txt")
        with open(requirement1, "w") as fd:
            fd.write(
                """
inmanta-module-std

                """
            )
        with open(requirement2, "w") as fd:
            fd.write(
                """
inmanta-module-net

inmanta-module-ip
                """
            )
        with open(constraint1, "w") as fd:
            fd.write(
                """
inmanta-module-std
                """
            )
        with open(constraint2, "w") as fd:
            fd.write(
                """

inmanta-module-ip
inmanta-module-net


                """
            )
        caplog.clear()
        tmpvenv_active_inherit._run_pip_install_command(
            python_path=env.process_env.python_path,
            constraints_files=[constraint1, constraint2],
            requirements_files=[requirement1, requirement2],
        )

        assert all(record.name == "inmanta.pip" for record in caplog.records)
        python_path: str = tmpvenv_active_inherit.python_path
        assert (
            f"""
Content of requirements files:
    {requirement1}:
        inmanta-module-std
    {requirement2}:
        inmanta-module-net
        inmanta-module-ip
Content of constraints files:
    {constraint1}:
        inmanta-module-std
    {constraint2}:
        inmanta-module-ip
        inmanta-module-net
Pip command: {python_path} -m pip install -c {constraint1} -c {constraint2} -r {requirement1} -r {requirement2}
""".strip()
            in caplog.messages
        )



"""
    Copyright 2023 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
import datetime
import logging
import uuid
from collections import abc, defaultdict
from typing import Dict, List, Literal, Optional, Set, cast

import asyncpg
import asyncpg.connection
import asyncpg.exceptions
import pydantic

from inmanta import const, data
from inmanta.const import ResourceState
from inmanta.data import APILIMIT, AVAILABLE_VERSIONS_TO_KEEP, ENVIRONMENT_AGENT_TRIGGER_METHOD, PURGE_ON_DELETE, InvalidSort
from inmanta.data.dataview import DesiredStateVersionView
from inmanta.data.model import (
    DesiredStateVersion,
    PromoteTriggerMethod,
    ResourceDiff,
    ResourceIdStr,
    ResourceMinimal,
    ResourceVersionIdStr,
)
from inmanta.protocol import handle, methods, methods_v2
from inmanta.protocol.common import ReturnValue, attach_warnings
from inmanta.protocol.exceptions import BadRequest, BaseHttpException, NotFound, ServerError
from inmanta.resources import Id
from inmanta.server import (
    SLICE_AGENT_MANAGER,
    SLICE_AUTOSTARTED_AGENT_MANAGER,
    SLICE_DATABASE,
    SLICE_ORCHESTRATION,
    SLICE_RESOURCE,
    SLICE_TRANSPORT,
)
from inmanta.server import config as opt
from inmanta.server import diff, protocol
from inmanta.server.agentmanager import AgentManager, AutostartedAgentManager
from inmanta.server.services.resourceservice import ResourceService
from inmanta.server.validate_filter import InvalidFilter
from inmanta.types import Apireturn, JsonType, PrimitiveTypes

LOGGER = logging.getLogger(__name__)

PERFORM_CLEANUP: bool = True
# Kill switch for cleanup, for use when working with historical data


class CrossResourceSetDependencyError(Exception):
    def __init__(self, resource_id1: ResourceIdStr, resource_id2: ResourceIdStr) -> None:
        """
        Raised when a cross-resource set dependency was detected between the resource with id
        resource_id1 and resource_id2.
        """
        self.resource_id1 = resource_id1
        self.resource_id2 = resource_id2
        super().__init__(self.get_error_message())

    def get_error_message(self) -> str:
        return (
            f"A dependency exists between resources {self.resource_id1} and {self.resource_id2}, but they belong to"
            f" different resource sets."
        )


class ResourceSetValidator:
    def __init__(self, resources: abc.Set[data.Resource]) -> None:
        self.resources = resources
        self.rid_to_resource_set = {res.resource_id: res.resource_set for res in self.resources}

    def _is_cross_resource_set_dependency(self, res: data.Resource, rid_dependency: ResourceIdStr) -> bool:
        """
        Return True iff the dependency between resource res and the resource with id rid_dependency is a cross-resource set
        dependency.
        """
        if res.resource_set is None:
            # Resource is in shared resource set.
            return False
        if rid_dependency not in self.rid_to_resource_set:
            # A partial compile was done and we have a dependency on a resource in another resource set
            # that is not part of the partial compile.
            return True
        resource_set_dep = self.rid_to_resource_set[rid_dependency]
        if resource_set_dep is None:
            # Dependency towards shared resource set.
            return False
        return res.resource_set != resource_set_dep

    def ensure_no_cross_resource_set_dependencies(self) -> None:
        """
        This method raises a CrossResourceSetDependencyError when a resource in self.resources that belongs to a non-shared
        resource set has a dependency (requires/provides) on another resource that belongs to a different non-shared resource
        set.
        """
        for res in self.resources:
            # It's sufficient to only check the requires relationship. The provides
            # relationship is always a subset of the reverse relationship (provides relationship).
            # The provides relationship only contains the cross-agent dependencies.
            for req in res.get_requires():
                if self._is_cross_resource_set_dependency(res, req):
                    raise CrossResourceSetDependencyError(res.resource_id, req)

    def has_cross_resource_set_dependency(self) -> bool:
        """
        Return True iff a cross resource set dependency exists between the resources in self.resources.
        """
        try:
            self.ensure_no_cross_resource_set_dependencies()
        except CrossResourceSetDependencyError:
            return True
        else:
            return False


class PartialUpdateMerger:
    """
    Class that contains the functionality to merge the shared resources and resources, present in a resource set that is updated
    by the partial compile, together with the resources from the corresponding resources sets in the old version of the model.
    """

    def __init__(
        self,
        env_id: uuid.UUID,
        base_version: int,
        version: int,
        rids_in_partial_compile: abc.Set[ResourceIdStr],
        updated_resource_sets: abc.Set[str],
        deleted_resource_sets: abc.Set[str],
        updated_and_shared_resources_old: abc.Mapping[ResourceIdStr, data.Resource],
        rids_deleted_resource_sets: abc.Set[ResourceIdStr],
    ) -> None:
        """
        :param env_id: The id of the environment for which a partial compile is being done.
        :param base_version: The source version on which the partial compile in based.
        :param version: The version of the new configuration model created by this partial compile.
        :param rids_in_partial_compile: The ids of the resource that are part of the partial compile.
        :param updated_resource_sets: The names of the resource sets that are updated by the partial compile.
        :param deleted_resource_sets: The names of the resource sets that are deleted by the partial compile.
        :param updated_and_shared_resources_old: A dictionary that contains all the resources in base_version that belong
                                                 to a resource set in updated_resource_sets or to the shared resource set.
        :param rids_deleted_resource_sets: The ids of the resources that are in the base_version and that are deleted by this
                                           partial compile.
        """
        self.env_id = env_id
        self.base_version = base_version
        self.version = version
        self.rids_in_partial_compile = rids_in_partial_compile
        self.updated_resource_sets = updated_resource_sets
        self.deleted_resource_sets = deleted_resource_sets
        self.updated_and_shared_resources_old = updated_and_shared_resources_old
        self.non_shared_resources_in_partial_update_old: abc.Mapping[ResourceIdStr, data.Resource] = {
            rid: r for rid, r in self.updated_and_shared_resources_old.items() if r.resource_set is not None
        }
        self.shared_resources_old: abc.Mapping[ResourceIdStr, data.Resource] = {
            rid: r for rid, r in self.updated_and_shared_resources_old.items() if r.resource_set is None
        }
        self.rids_deleted_resource_sets = rids_deleted_resource_sets

    @classmethod
    async def create(
        cls,
        env_id: uuid.UUID,
        base_version: int,
        version: int,
        rids_in_partial_compile: abc.Set[ResourceIdStr],
        updated_resource_sets: abc.Set[str],
        deleted_resource_sets: abc.Set[str],
        connection: Optional[asyncpg.connection.Connection] = None,
    ) -> "PartialUpdateMerger":
        """
        A replacement constructor method for this class. This method is used to work around the limitation that no async
        calls can be done in a constructor. See docstring real constructor for meaning of arguments.
        """
        updated_and_shared_resources_old: abc.Mapping[
            ResourceIdStr, data.Resource
        ] = await data.Resource.get_resources_in_resource_sets(
            environment=env_id,
            version=base_version,
            resource_sets=updated_resource_sets,
            include_shared_resources=True,
            connection=connection,
        )
        rids_deleted_resource_sets: abc.Set[ResourceIdStr] = set(
            rid
            for rid in (
                await data.Resource.get_resources_in_resource_sets(
                    environment=env_id,
                    version=base_version,
                    resource_sets=deleted_resource_sets,
                    connection=connection,
                )
            ).keys()
        )
        return PartialUpdateMerger(
            env_id,
            base_version,
            version,
            rids_in_partial_compile,
            updated_resource_sets,
            deleted_resource_sets,
            updated_and_shared_resources_old,
            rids_deleted_resource_sets,
        )

    def merge_updated_and_shared_resources(
        self, updated_and_shared_resources: abc.Sequence[data.Resource]
    ) -> Dict[ResourceIdStr, data.Resource]:
        """
         Separates named resource sets from the shared resource set and expands the shared set with the shared resources in
         the previous model version.

        :param updated_and_shared_resources: The resources that are part of the partial compile.
        :returns: The subset of resources in the new version of the configuration model that belong to the shared resource set
                  or a resource set that is updated by this partial compile.
        """
        shared_resources = {r.resource_id: r for r in updated_and_shared_resources if r.resource_set is None}
        updated_resources = {r.resource_id: r for r in updated_and_shared_resources if r.resource_set is not None}
        shared_resources_merged = {r.resource_id: r for r in self._merge_shared_resources(shared_resources)}
        result = {**updated_resources, **shared_resources_merged}
        self._validate_constraints(result)
        return result

    def _validate_constraints(self, new_updated_and_shared_resources: abc.Mapping[ResourceIdStr, data.Resource]) -> None:
        """
        Validate whether the new updated and shared resources that results from the merging the old version of the model
        with resources of the partial compile, are compliant with the constraints of a partial compile.

        :param new_updated_and_shared_resources: The resources that have to be validated.
        """
        for res_id, res in new_updated_and_shared_resources.items():
            if res.resource_id not in self.updated_and_shared_resources_old:
                continue
            matching_resource_old_model = self.updated_and_shared_resources_old[res.resource_id]

            if res.resource_set != matching_resource_old_model.resource_set:
                raise BadRequest(
                    f"A partial compile cannot migrate resources: trying to move {res.resource_id} from resource set"
                    f" {matching_resource_old_model.resource_set} to {res.resource_set}."
                )

            if res.resource_set is None and res.attribute_hash != matching_resource_old_model.attribute_hash:
                raise BadRequest(f"Resource ({res.resource_id}) without a resource set cannot be updated via a partial compile")

            resource_set_validator = ResourceSetValidator(set(new_updated_and_shared_resources.values()))
            try:
                resource_set_validator.ensure_no_cross_resource_set_dependencies()
            except CrossResourceSetDependencyError as e:
                raise BadRequest(e.get_error_message())

    def _merge_shared_resources(self, shared_resources_new: Dict[ResourceIdStr, data.Resource]) -> abc.Sequence[data.Resource]:
        """
        Merge the set of shared resources present in the old version of the model together with the set of shared resources
        present in the partial compile.

        :param shared_resources_new: The set of shared resources present in the partial compile.
        :returns: The set of shared resources that should be present in the new version of the model.
        """
        all_rids_shared_resources = set(self.shared_resources_old.keys()) | set(shared_resources_new.keys())
        result = []
        for rid_shared_resource in all_rids_shared_resources:
            if rid_shared_resource in shared_resources_new and rid_shared_resource in self.shared_resources_old:
                # Merge requires/provides shared resource
                old_shared_resource = self.shared_resources_old[rid_shared_resource]
                new_shared_resource = shared_resources_new[rid_shared_resource]
                res = self._merge_requires_and_provides_of_shared_resource(old_shared_resource, new_shared_resource)
            elif rid_shared_resource in shared_resources_new:
                # New shared resource in partial compile
                res = shared_resources_new[rid_shared_resource]
            else:
                # Old shared resource not referenced by partial compile
                res_old = self.shared_resources_old[rid_shared_resource]
                res = res_old.copy_for_partial_compile(new_version=self.version)
                res = self._clean_requires_provides_old_shared_resource(res)
            result.append(res)
        return result

    def _should_keep_dependency_old_shared_resources(self, rid_dependency: ResourceIdStr) -> bool:
        """
        Return True iff the given dependency present in a shared resource from the base version should be retained
        in the new version of the model.
        """
        if rid_dependency in self.rids_deleted_resource_sets:
            # Resource belongs to a deleted resource set
            return False
        if rid_dependency in self.non_shared_resources_in_partial_update_old:
            # If this dependency is still present in the new version of the model, this dependency
            # will be present in the resources that are part of the partial compile.
            return False
        return True

    def _clean_requires_provides_old_shared_resource(self, resource: data.Resource) -> data.Resource:
        """
        Cleanup the requires/provides relationship for shared resources that are not present in the partial compile
        and that were copied from the old version of the model.
        """
        resource.attributes["requires"] = [
            rid for rid in resource.attributes["requires"] if self._should_keep_dependency_old_shared_resources(rid)
        ]
        resource.provides = [rid for rid in resource.provides if self._should_keep_dependency_old_shared_resources(rid)]
        return resource

    def _merge_requires_and_provides_of_shared_resource(self, old: data.Resource, new: data.Resource) -> data.Resource:
        """
        Update the requires and provides relationship of `new` to make it consistent with the new version of the model.

        :param old: The shared resource present in the old version of the model.
        :param new: The shared resource part of the incremental compile.
        """
        new.provides = list(self._merge_dependencies_shared_resource(old.provides, new.provides))
        new.attributes["requires"] = self._merge_dependencies_shared_resource(old.get_requires(), new.get_requires())
        return new

    def _merge_dependencies_shared_resource(
        self, old_deps: abc.Sequence[ResourceIdStr], new_deps: abc.Sequence[ResourceIdStr]
    ) -> abc.Sequence[ResourceIdStr]:
        """
        Merge the dependencies for a certain shared resource together to make it consistent with the new version of the model.

        :param old_deps: The set of dependencies present in the old version of the shared resource.
        :param new_deps: The set of dependencies present in the shared resource that is part of the partial compile.
        """
        old_deps_cleaned: abc.Set[ResourceIdStr] = {
            dep for dep in old_deps if self._should_keep_dependency_old_shared_resources(dep)
        }
        return list(old_deps_cleaned | set(new_deps))

    async def merge_unknowns(
        self, unknowns_in_partial_compile: abc.Sequence[data.UnknownParameter]
    ) -> abc.Sequence[data.UnknownParameter]:
        """
        Merge all relevant, unresolved unknowns from the old version of the model together with the unknowns
        of the partial compile.
        """
        old_unresolved_unknowns_to_keep = [
            uk.copy(self.version)
            for uk in await data.UnknownParameter.get_unknowns_to_copy_in_partial_compile(
                environment=self.env_id,
                source_version=self.base_version,
                updated_resource_sets=self.updated_resource_sets,
                deleted_resource_sets=self.deleted_resource_sets,
                rids_in_partial_compile=self.rids_in_partial_compile,
            )
        ]
        return [*old_unresolved_unknowns_to_keep, *unknowns_in_partial_compile]


class OrchestrationService(protocol.ServerSlice):
    """Resource Manager service"""

    agentmanager_service: "AgentManager"
    autostarted_agent_manager: AutostartedAgentManager
    resource_service: ResourceService

    def __init__(self) -> None:
        super(OrchestrationService, self).__init__(SLICE_ORCHESTRATION)

    def get_dependencies(self) -> List[str]:
        return [SLICE_RESOURCE, SLICE_AGENT_MANAGER, SLICE_DATABASE]

    def get_depended_by(self) -> List[str]:
        return [SLICE_TRANSPORT]

    async def prestart(self, server: protocol.Server) -> None:
        await super().prestart(server)
        self.agentmanager_service = cast("AgentManager", server.get_slice(SLICE_AGENT_MANAGER))
        self.autostarted_agent_manager = cast(AutostartedAgentManager, server.get_slice(SLICE_AUTOSTARTED_AGENT_MANAGER))
        self.resource_service = cast(ResourceService, server.get_slice(SLICE_RESOURCE))

    async def start(self) -> None:
        if PERFORM_CLEANUP:
            self.schedule(self._purge_versions, opt.server_purge_version_interval.get(), cancel_on_stop=False)
            self.add_background_task(self._purge_versions())
        await super().start()

    async def _purge_versions(self) -> None:
        """
        Purge versions from the database
        """
        # TODO: move to data and use queries for delete
        envs = await data.Environment.get_list(halted=False)
        for env_item in envs:
            # get available versions
            n_versions = await env_item.get(AVAILABLE_VERSIONS_TO_KEEP)
            assert isinstance(n_versions, int)
            versions = await data.ConfigurationModel.get_list(environment=env_item.id)
            if len(versions) > n_versions:
                LOGGER.info("Removing %s available versions from environment %s", len(versions) - n_versions, env_item.id)
                version_dict = {x.version: x for x in versions}
                delete_list = sorted(version_dict.keys())
                delete_list = delete_list[:-n_versions]

                for v in delete_list:
                    await version_dict[v].delete_cascade()

        # Cleanup old agents from agent table in db
        await data.Agent.clean_up()

    @handle(methods.list_versions, env="tid")
    async def list_version(self, env: data.Environment, start: Optional[int] = None, limit: Optional[int] = None) -> Apireturn:
        if (start is None and limit is not None) or (limit is None and start is not None):
            raise ServerError("Start and limit should always be set together.")

        if start is None or limit is None:
            start = 0
            limit = data.APILIMIT

        if limit > APILIMIT:
            raise BadRequest(f"limit parameter can not exceed {APILIMIT}, got {limit}.")

        models = await data.ConfigurationModel.get_versions(env.id, start, limit)
        count = len(models)

        d = {
            "versions": models,
            "start": start,
            "limit": limit,
            "count": count,
        }

        return 200, d

    @handle(methods.get_version, version_id="id", env="tid")
    async def get_version(
        self,
        env: data.Environment,
        version_id: int,
        include_logs: Optional[bool] = None,
        log_filter: Optional[str] = None,
        limit: Optional[int] = 0,
    ) -> Apireturn:
        version = await data.ConfigurationModel.get_version(env.id, version_id)
        if version is None:
            return 404, {"message": "The given configuration model does not exist yet."}

        resources = await data.Resource.get_resources_for_version(env.id, version_id, no_obj=True)
        if resources is None:
            return 404, {"message": "The given configuration model does not exist yet."}

        if limit is None:
            limit = APILIMIT
        elif limit > APILIMIT:
            raise BadRequest(
                f"limit parameter can not exceed {APILIMIT}, got {limit}."
                f" To retrieve more entries, use /api/v2/resource_actions"
            )

        resources_out: List[JsonType] = []
        d = {"model": version, "resources": resources_out}
        resource_action_lookup: Dict[ResourceVersionIdStr, List[data.ResourceAction]] = {}

        for res_dict in resources:
            resources_out.append(res_dict)
            if bool(include_logs):
                actions: List[data.ResourceAction] = []
                res_dict["actions"] = actions
                resource_action_lookup[res_dict["resource_version_id"]] = actions

        if include_logs:
            # get all logs, unsorted
            all_logs = await data.ResourceAction.get_logs_for_version(env.id, version_id, log_filter, limit)
            for log in all_logs:
                for resource_version_id in log.resource_version_ids:
                    resource_action_lookup[resource_version_id].append(log)

        d["unknowns"] = await data.UnknownParameter.get_list(environment=env.id, version=version_id)

        return 200, d

    @handle(methods.delete_version, version_id="id", env="tid")
    async def delete_version(self, env: data.Environment, version_id: int) -> Apireturn:
        version = await data.ConfigurationModel.get_version(env.id, version_id)
        if version is None:
            return 404, {"message": "The given configuration model does not exist yet."}

        await version.delete_cascade()
        return 200

    @handle(methods_v2.reserve_version, env="tid")
    async def reserve_version(self, env: data.Environment) -> int:
        return await env.get_next_version()

    def _create_dao_resources_from_api_resources(
        self,
        env_id: uuid.UUID,
        resources: List[JsonType],
        resource_state: Dict[ResourceIdStr, Literal[ResourceState.available, ResourceState.undefined]],
        resource_sets: Dict[ResourceIdStr, Optional[str]],
        set_version: Optional[int] = None,
    ) -> Dict[ResourceIdStr, data.Resource]:
        """
        This method converts the resources sent to the put_version or put_partial endpoint to dao Resource objects.
        The resulting resource objects will have their provides set up correctly for cross agent dependencies
        and the version field of these resources will be set to set_version if provided.

        An exception will be raised when the one of the following constraints is not satisfied:
            * A resource present in the resource_sets parameter is not present in the resources dictionary.
            * The dependency graph of the provided resources is not closed.
        """
        rid_to_resource = {}
        # The content of the requires attribute for all the resources
        all_requires: set[ResourceIdStr] = set()
        # list of all resources which have a cross agent dependency, as a tuple, (dependant,requires)
        cross_agent_dep: list[tuple[data.Resource, Id]] = []
        for res_dict in resources:
            # Verify that the version field and the version in the resource version id field match
            version_part_of_resource_id = Id.parse_id(res_dict["id"]).version
            if "version" in res_dict and res_dict["version"] != version_part_of_resource_id:
                raise BadRequest(
                    f"Invalid resource: The version in the id field ({res_dict['id']}) doesn't match the version in the"
                    f" version field ({res_dict['version']})."
                )
            res_obj = data.Resource.new(env_id, res_dict["id"])
            # Populate status field
            if res_obj.resource_id in resource_state:
                res_obj.status = const.ResourceState[resource_state[res_obj.resource_id]]
            # Populate resource_set field
            if res_obj.resource_id in resource_sets:
                res_obj.resource_set = resource_sets[res_obj.resource_id]

            # Populate attributes field of resources
            attributes = {}
            for field, value in res_dict.items():
                if field not in {"id", "version"}:
                    attributes[field] = value
            res_obj.attributes = attributes
            res_obj.make_hash()

            # Update the version fields
            if set_version is not None:
                res_obj.model = set_version

            # find cross agent dependencies
            agent = res_obj.agent
            if "requires" not in attributes:
                LOGGER.warning("Received resource without requires attribute (%s)", res_obj.resource_id)
            else:
                # Collect all requires as resource_ids instead of resource version ids
                cleaned_requires = []
                for req in attributes["requires"]:
                    rid = Id.parse_id(req)
                    all_requires.add(rid.resource_str())
                    if rid.get_agent_name() != agent:
                        # it is a CAD
                        cross_agent_dep.append((res_obj, rid))
                    cleaned_requires.append(rid.resource_str())
                attributes["requires"] = cleaned_requires

            rid_to_resource[res_obj.resource_id] = res_obj

        # hook up all CADs
        for f, t in cross_agent_dep:
            res_obj = rid_to_resource[t.resource_str()]
            res_obj.provides.append(f.resource_id)

        rids = set(rid_to_resource.keys())

        # Sanity checks
        superfluous_ids = set(resource_sets.keys()) - rids
        if superfluous_ids:
            raise BadRequest(
                "The following resource ids provided in the resource_sets parameter are not present "
                f"in the resources list: {', '.join(superfluous_ids)}"
            )
        if not all_requires.issubset(rids):
            raise BadRequest(
                "The model should have a dependency graph that is closed and no dangling dependencies:"
                f" {all_requires - rids}"
            )

        return rid_to_resource

    def _get_skipped_for_undeployable(
        self, resources: abc.Sequence[data.Resource], undeployable_ids: abc.Sequence[ResourceIdStr]
    ) -> abc.Sequence[ResourceIdStr]:
        """
        Return the resources that are skipped_for_undeployable given the full set of resources and
        the resource ids of the resources that are undeployable.

        :param resources: All resources in the model.
        :param undeployable_ids: The ids of the resource that are undeployable.
        """
        # Build up provides tree
        provides_tree: Dict[ResourceIdStr, List[ResourceIdStr]] = defaultdict(lambda: [])
        for r in resources:
            if "requires" in r.attributes:
                for req in r.attributes["requires"]:
                    req_id = Id.parse_id(req)
                    provides_tree[req_id.resource_str()].append(r.resource_id)
        # Find skipped for undeployables
        work = list(undeployable_ids)
        skippeable: Set[ResourceIdStr] = set()
        while len(work) > 0:
            current = work.pop()
            if current in skippeable:
                continue
            skippeable.add(current)
            work.extend(provides_tree[current])
        return list(skippeable - set(undeployable_ids))

    async def _put_version(
        self,
        env: data.Environment,
        version: int,
        rid_to_resource: Dict[ResourceIdStr, data.Resource],
        unknowns: abc.Sequence[data.UnknownParameter],
        version_info: Optional[JsonType] = None,
        resource_sets: Optional[Dict[ResourceIdStr, Optional[str]]] = None,
        partial_base_version: Optional[int] = None,
        removed_resource_sets: Optional[List[str]] = None,
        *,
        connection: asyncpg.connection.Connection,
    ) -> None:
        """
        :param rid_to_resource: This parameter should contain all the resources when a full compile is done.
                                When a partial compile is done, it should contain all the resources that belong to the
                                updated resource sets or the shared resource sets. This method updates this object with
                                purge-on-delete resources.
        :param unknowns: This parameter should contain all the unknowns for all the resources in the new version of the model.
                         Also the unknowns for resources that are not present in rid_to_resource.
        :param partial_base_version: When a partial compile is done, this parameter contains the version of the
                                     configurationmodel this partial compile was based on. Otherwise this parameter should be
                                     None.
        :param removed_resource_sets: When a partial compile is done, this parameter should indicate the names of the resource
                                      sets that are removed by the partial compile. When no resource sets are removed by
                                      a partial compile or when a full compile is done, this parameter can be set to None.

        Pre-conditions:
            * The requires and provides relationships of the resources in rid_to_resource must be set correctly. For a
              partial compile, this means it is assumed to be valid with respect to all absolute constraints that apply to
              partial compiles. Constraints that are relative to the base version will be verified by this method.
            * When a partial compile was done, all resources in rid_to_resource must meet the constraints of a partial compile.
            * The resource sets defined in the removed_resource_sets argument must not overlap with the resource sets present
              in the resource_sets argument.

        This method adds resources to the model that are no longer present in the new version of the model and had the
        purge_on_delete flag set to true. This method makes sure that the requires and provides relationships of these
        additional resources are set correctly. This dependency wiring is also safe for a partial compile because
        a partial compile cannot delete shared resources and because resources that belong to a non-shared resource set cannot
        reference resources in another non-shared resource set. That way all dependencies of the deleted resources are present
        in the rid_to_resource parameter.

        When a partial compile is done, the undeployable and skipped_for_undeployable resources of a configurationmodel are
        copied from the old version to the new version. This operation is safe because the only resources missing from
        rid_to_resource are resources that belong to an unchanged, non-shared resource set. Those resources can only have
        cross resource set dependencies in a non-shared resource set and the latter resource set cannot be changed by a partial
        compile.

        Validations done by this method:
            * In case of a full export: Checks whether this version has any requires-provides across resource sets and
                                        sets the is_suitable_for_partial_compiles field appropriately, indicating whether
                                        this version is eligible to be used as a base version for a future partial compile.
            * In case of a partial export: Verifies that no resources moved resource sets.
        """
        is_partial_update = partial_base_version is not None

        if resource_sets is None:
            resource_sets = {}

        if removed_resource_sets is None:
            removed_resource_sets = []

        if version > env.last_version:
            raise BadRequest(
                f"The version number used is {version} "
                f"which is higher than the last outstanding reservation {env.last_version}"
            )
        if version <= 0:
            raise BadRequest(f"The version number used ({version}) is not positive")

        for r in rid_to_resource.values():
            if r.model != version:
                raise BadRequest(
                    f"The resource version of resource {r.resource_version_id} does not match the version argument "
                    f"(version: {version})"
                )

        for rid_name in resource_sets.keys():
            try:
                Id.parse_id(rid_name)
            except Exception as e:
                raise BadRequest("Invalid resource id in resource set: %s" % str(e))

        started = datetime.datetime.now().astimezone()

        resource_set_validator = ResourceSetValidator(set(rid_to_resource.values()))
        undeployable_ids: abc.Sequence[ResourceIdStr] = [
            res.resource_id for res in rid_to_resource.values() if res.status in const.UNDEPLOYABLE_STATES
        ]
        async with connection.transaction():
            try:
                if is_partial_update:
                    # Make mypy happy
                    assert partial_base_version is not None
                    cm = await data.ConfigurationModel.create_for_partial_compile(
                        env_id=env.id,
                        version=version,
                        # When a partial compile is done, the total will be updated in cm.recalculate_total()
                        # with all the resources that belong to a resource set that was not updated.
                        total=len(rid_to_resource),
                        version_info=version_info,
                        undeployable=undeployable_ids,
                        skipped_for_undeployable=sorted(
                            self._get_skipped_for_undeployable(list(rid_to_resource.values()), undeployable_ids)
                        ),
                        partial_base=partial_base_version,
                        rids_in_partial_compile=set(rid_to_resource.keys()),
                        connection=connection,
                    )
                else:
                    cm = data.ConfigurationModel(
                        environment=env.id,
                        version=version,
                        date=datetime.datetime.now().astimezone(),
                        total=len(rid_to_resource),
                        version_info=version_info,
                        undeployable=undeployable_ids,
                        skipped_for_undeployable=sorted(
                            self._get_skipped_for_undeployable(list(rid_to_resource.values()), undeployable_ids)
                        ),
                        is_suitable_for_partial_compiles=not resource_set_validator.has_cross_resource_set_dependency(),
                    )
                    await cm.insert(connection=connection)
            except asyncpg.exceptions.UniqueViolationError:
                raise ServerError("The given version is already defined. Versions should be unique.")

            all_ids: set[Id] = set(Id.parse_id(rid, version) for rid in rid_to_resource.keys())
            if is_partial_update:
                # Make mypy happy
                assert partial_base_version is not None
                # This dict maps a resource id to its resource set for unchanged resource sets.
                rids_unchanged_resource_sets: dict[
                    ResourceIdStr, str
                ] = await data.Resource.copy_resources_from_unchanged_resource_set(
                    environment=env.id,
                    source_version=partial_base_version,
                    destination_version=version,
                    updated_resource_sets=set(sr for sr in resource_sets.values() if sr is not None),
                    deleted_resource_sets=set(removed_resource_sets),
                    connection=connection,
                )
                resources_that_moved_resource_sets = rids_unchanged_resource_sets.keys() & rid_to_resource.keys()
                if resources_that_moved_resource_sets:
                    msg = (
                        "The following Resource(s) cannot be migrated to a different resource set using a partial compile, "
                        "a full compile is necessary for this process:\n"
                    )
                    msg += "\n".join(
                        f"    {rid} moved from {rids_unchanged_resource_sets[rid]} to {resource_sets[rid]}"
                        for rid in resources_that_moved_resource_sets
                    )

                    raise BadRequest(msg)
                all_ids |= {Id.parse_id(rid, version) for rid in rids_unchanged_resource_sets.keys()}

            purge_on_delete_resources: Dict[ResourceIdStr, data.Resource] = await self._get_resources_for_purge_on_delete(
                env=env,
                version=version,
                rid_to_resource=rid_to_resource,
                all_resource_ids=[i.resource_str() for i in all_ids],
                version_info=version_info,
                connection=connection,
            )
            rid_to_resource.update(purge_on_delete_resources)
            all_ids |= {Id.parse_id(rid, version) for rid in purge_on_delete_resources.keys()}

            await data.Resource.insert_many(list(rid_to_resource.values()), connection=connection)
            await cm.recalculate_total(connection=connection)

            await data.UnknownParameter.insert_many(unknowns, connection=connection)

            all_agents: abc.Set[str] = {res.agent for res in rid_to_resource.values()}
            for agent in all_agents:
                await self.agentmanager_service.ensure_agent_registered(env, agent, connection=connection)

            # Don't log ResourceActions without resource_version_ids, because
            # no API call exists to retrieve them.
            all_rvids = [i.resource_version_str() for i in all_ids]
            if all_rvids:
                now = datetime.datetime.now().astimezone()
                log_line = data.LogLine.log(logging.INFO, "Successfully stored version %(version)d", version=version)
                self.resource_service.log_resource_action(env.id, list(all_rvids), logging.INFO, now, log_line.msg)
                ra = data.ResourceAction(
                    environment=env.id,
                    version=version,
                    resource_version_ids=all_rvids,
                    action_id=uuid.uuid4(),
                    action=const.ResourceAction.store,
                    started=started,
                    finished=now,
                    messages=[log_line],
                )
                await ra.insert(connection=connection)

        LOGGER.debug("Successfully stored version %d", version)

    async def _get_resources_for_purge_on_delete(
        self,
        env: data.Environment,
        version: int,
        rid_to_resource: abc.Mapping[ResourceIdStr, data.Resource],
        all_resource_ids: abc.Sequence[ResourceIdStr],
        version_info: Optional[JsonType] = None,
        *,
        connection: asyncpg.connection.Connection,
    ) -> dict[ResourceIdStr, data.Resource]:
        """
        Return a dictionary of resources that were present in the old version of the model but that no longer exist in this
        version (in rid_to_resource) and had the purge_on_delete flag set to true. The returned resources will have the purged
        attribute set to True and their requires/provides relationships inverted.

        :param env: The environment we are discovering purged resources for.
        :param version: The new version of the configuration model in which the purge_on_delete resources should be added.
        :param rid_to_resource: When a full compile is done, this dictionary contains an entry for all resources that are
                                part of the full compile. When a partial compile is done, this dictionary contains an entry
                                for each resource in the new version of the model that belongs to an updated or shared
                                resource set.
        :param all_resource_ids: All the resource ids of all the resources present in the new version of the model. When
                                 a partial compile is done, this sequence also includes the resource ids of resources that were
                                 not updated by the partial compile.
        """

        def safe_get(input: object, key: str, default: object) -> object:
            if not isinstance(input, dict):
                return default
            if key not in input:
                return default
            return input[key]

        rid_to_resource = dict(rid_to_resource)

        # detect failed compiles
        metadata: object = safe_get(version_info, const.EXPORT_META_DATA, {})
        compile_state = safe_get(metadata, const.META_DATA_COMPILE_STATE, "")
        failed = compile_state == const.Compilestate.failed

        result = {}
        if not failed and (await env.get(PURGE_ON_DELETE)):
            # search for deleted resources (purge_on_delete)
            resources_to_purge: List[data.Resource] = await data.Resource.get_deleted_resources(
                env.id, version, all_resource_ids, connection=connection
            )

            previous_requires = {}
            for res in resources_to_purge:
                LOGGER.warning("Purging %s, purged resource based on %s", res.resource_id, res.resource_version_id)

                attributes = res.attributes.copy()
                attributes["purged"] = True
                attributes["requires"] = []
                res_obj = data.Resource.new(
                    env.id,
                    resource_version_id=ResourceVersionIdStr("%s,v=%s" % (res.resource_id, version)),
                    attributes=attributes,
                )

                previous_requires[res_obj.resource_id] = res.attributes["requires"]
                rid_to_resource[res_obj.resource_id] = res_obj
                result[res_obj.resource_id] = res_obj

            # invert dependencies on purges
            for res_id, requires in previous_requires.items():
                res_obj = rid_to_resource[res_id]
                for require in requires:
                    req_id = Id.parse_id(require)

                    if req_id.resource_str() in rid_to_resource:
                        req_res = rid_to_resource[req_id.resource_str()]
                        req_res.attributes["requires"].append(res_obj.resource_id)
                        if res_obj.agent != req_res.agent:
                            res_obj.provides.append(req_res.resource_id)
        return result

    async def _trigger_auto_deploy(
        self,
        env: data.Environment,
        version: int,
    ) -> None:
        """
        Triggers auto-deploy for stored resources. Must be called only after transaction that stores resources has been allowed
        to commit. If not respected, the auto deploy might work on stale data, likely resulting in resources hanging in the
        deploying state.
        """
        auto_deploy = await env.get(data.AUTO_DEPLOY)
        if auto_deploy:
            LOGGER.debug("Auto deploying version %d", version)
            push_on_auto_deploy = cast(bool, await env.get(data.PUSH_ON_AUTO_DEPLOY))
            agent_trigger_method_on_autodeploy = cast(str, await env.get(data.AGENT_TRIGGER_METHOD_ON_AUTO_DEPLOY))
            agent_trigger_method_on_autodeploy = const.AgentTriggerMethod[agent_trigger_method_on_autodeploy]
            await self.release_version(env, version, push_on_auto_deploy, agent_trigger_method_on_autodeploy)

    def _create_unknown_parameter_daos_from_api_unknowns(
        self, env_id: uuid.UUID, version: int, unknowns: Optional[List[Dict[str, PrimitiveTypes]]] = None
    ) -> List[data.UnknownParameter]:
        """
        Create UnknownParameter dao's from the unknowns dictionaries passed through the put_version() and put_partial API
        endpoint.
        """
        if not unknowns:
            return []
        result = []
        for uk in unknowns:
            if "resource" not in uk:
                uk["resource"] = ""
            if "metadata" not in uk:
                uk["metadata"] = {}
            unknown_parameter = data.UnknownParameter(
                resource_id=uk["resource"],
                name=uk["parameter"],
                source=uk["source"],
                environment=env_id,
                version=version,
                metadata=uk["metadata"],
            )
            result.append(unknown_parameter)
        return result

    async def _put_version_lock(self, env: data.Environment, *, shared: bool = False, connection: asyncpg.Connection) -> None:
        """
        Acquires a transaction-level advisory lock for concurrency control between put_version and put_partial.

        :param env: The environment to acquire the lock for.
        :param shared: If true, doesn't conflict with other shared locks, only with non-shared once.
        :param connection: The connection hosting the transaction for which to acquire a lock.
        """
        lock: str = "pg_advisory_xact_lock_shared" if shared else "pg_advisory_xact_lock"
        await connection.execute(
            # Advisory lock keys are only 32 bit (or a single 64 bit key), while a full uuid is 128 bit.
            # Since locking slightly too strictly at extremely low odds is acceptable, we only use a 32 bit subvalue
            # of the uuid. For uuid4, time_low is (despite the name) randomly generated. Since it is an unsigned
            # integer while Postgres expects a signed one, we shift it by 2**31.
            f"SELECT {lock}({const.PG_ADVISORY_KEY_PUT_VERSION}, {env.id.time_low - 2**31})"
        )

    @handle(methods.put_version, env="tid")
    async def put_version(
        self,
        env: data.Environment,
        version: int,
        resources: List[JsonType],
        resource_state: Dict[ResourceIdStr, Literal[ResourceState.available, ResourceState.undefined]],
        unknowns: List[Dict[str, PrimitiveTypes]],
        version_info: JsonType,
        compiler_version: Optional[str] = None,
        resource_sets: Optional[Dict[ResourceIdStr, Optional[str]]] = None,
    ) -> Apireturn:
        """
        :param unknowns: dict with the following structure
                            {
                             "resource": ResourceIdStr,
                             "parameter": str,
                             "source": str
                            }
        """
        if resource_sets is None:
            resource_sets = {}

        if not compiler_version:
            raise BadRequest("Older compiler versions are no longer supported, please update your compiler")

        unknowns_objs = self._create_unknown_parameter_daos_from_api_unknowns(env.id, version, unknowns)
        rid_to_resource = self._create_dao_resources_from_api_resources(
            env_id=env.id,
            resources=resources,
            resource_state=resource_state,
            resource_sets=resource_sets,
        )
        async with data.Resource.get_connection() as con:
            async with con.transaction():
                # Acquire a lock that conflicts with the lock acquired by put_partial but not with itself
                await self._put_version_lock(env, shared=True, connection=con)
                await self._put_version(
                    env, version, rid_to_resource, unknowns_objs, version_info, resource_sets, connection=con
                )
        await self._trigger_auto_deploy(env, version)
        return 200

    @handle(methods_v2.put_partial, env="tid")
    async def put_partial(
        self,
        env: data.Environment,
        resources: object,
        resource_state: Optional[Dict[ResourceIdStr, Literal[ResourceState.available, ResourceState.undefined]]] = None,
        unknowns: Optional[List[Dict[str, PrimitiveTypes]]] = None,
        version_info: Optional[JsonType] = None,
        resource_sets: Optional[Dict[ResourceIdStr, Optional[str]]] = None,
        removed_resource_sets: Optional[List[str]] = None,
    ) -> int:
        """
        :param unknowns: dict with the following structure
                    {
                     "resource": ResourceIdStr,
                     "parameter": str,
                     "source": str
                    }
        """
        if resource_state is None:
            resource_state = {}
        if unknowns is None:
            unknowns = []
        if resource_sets is None:
            resource_sets = {}
        if removed_resource_sets is None:
            removed_resource_sets = []

        try:
            pydantic.parse_obj_as(List[ResourceMinimal], resources)
        except pydantic.ValidationError:
            raise BadRequest(
                "Type validation failed for resources argument. "
                f"Expected an argument of type List[Dict[str, Any]] but received {resources}"
            )
        else:
            # Make mypy happy
            resources = cast(List[JsonType], resources)

        # validate resources before any side effects take place
        for r in resources:
            rid = Id.parse_id(r["id"])
            if rid.get_version() != 0:
                raise BadRequest("Resources for partial export should not contain version information")

        intersection: set[str] = set(resource_sets.values()).intersection(set(removed_resource_sets))
        if intersection:
            raise BadRequest(
                "Following resource sets are present in the removed resource sets and in the resources that are exported: "
                f"{intersection}"
            )

        async with data.Resource.get_connection() as con:
            async with con.transaction():
                # Acquire a lock that conflicts with itself and with the lock acquired by put_version
                await self._put_version_lock(env, shared=False, connection=con)

                # Only request a new version once the resource lock has been acquired to ensure a monotonic version history
                version: int = await env.get_next_version(connection=con)

                current_versions: abc.Sequence[data.ConfigurationModel] = await data.ConfigurationModel.get_versions(
                    env.id, limit=1
                )
                if not current_versions:
                    raise BadRequest("A partial export requires a base model but no versions have been exported yet.")

                base_model = current_versions[0]
                base_version: int = base_model.version
                if not base_model.is_suitable_for_partial_compiles:
                    resources_in_base_version = await data.Resource.get_resources_for_version(env.id, base_version)
                    resource_set_validator = ResourceSetValidator(set(resources_in_base_version))
                    try:
                        resource_set_validator.ensure_no_cross_resource_set_dependencies()
                    except CrossResourceSetDependencyError as e:
                        raise BadRequest(
                            f"Base version {base_version} is not suitable for a partial compile. {e.get_error_message()}"
                        )
                    else:
                        # This should never happen
                        LOGGER.warning(
                            "Base version %d was marked as not suitable for partial compiles, but no cross resource set"
                            " dependencies were found.",
                            base_version,
                        )

                rid_to_resource: Dict[ResourceIdStr, data.Resource] = self._create_dao_resources_from_api_resources(
                    env_id=env.id,
                    resources=resources,
                    resource_state=resource_state,
                    resource_sets=resource_sets,
                    set_version=version,
                )

                updated_resource_sets: abc.Set[str] = set(sr_name for sr_name in resource_sets.values() if sr_name is not None)
                partial_update_merger = await PartialUpdateMerger.create(
                    env_id=env.id,
                    base_version=base_version,
                    version=version,
                    rids_in_partial_compile=set(rid_to_resource.keys()),
                    updated_resource_sets=updated_resource_sets,
                    deleted_resource_sets=set(removed_resource_sets),
                    connection=con,
                )

                # add shared resources
                merged_resources = partial_update_merger.merge_updated_and_shared_resources(list(rid_to_resource.values()))

                await data.Code.copy_versions(env.id, base_version, version, connection=con)

                merged_unknowns = await partial_update_merger.merge_unknowns(
                    unknowns_in_partial_compile=self._create_unknown_parameter_daos_from_api_unknowns(env.id, version, unknowns)
                )

                await self._put_version(
                    env,
                    version,
                    merged_resources,
                    merged_unknowns,
                    version_info,
                    resource_sets,
                    partial_base_version=base_version,
                    removed_resource_sets=removed_resource_sets,
                    connection=con,
                )
        await self._trigger_auto_deploy(env, version)
        return version

    @handle(methods.release_version, version_id="id", env="tid")
    async def release_version(
        self,
        env: data.Environment,
        version_id: int,
        push: bool,
        agent_trigger_method: Optional[const.AgentTriggerMethod] = None,
        *,
        connection: Optional[asyncpg.connection.Connection] = None,
    ) -> Apireturn:
        model = await data.ConfigurationModel.get_version(env.id, version_id, connection=connection)
        if model is None:
            return 404, {"message": "The request version does not exist."}

        # Already mark undeployable resources as deployed to create a better UX (change the version counters)
        undep = await model.get_undeployable()
        now = datetime.datetime.now().astimezone()

        if undep:
            undep_ids = [ResourceVersionIdStr(rid + ",v=%s" % version_id) for rid in undep]
            # not checking error conditions
            await self.resource_service.resource_action_update(
                env,
                undep_ids,
                action_id=uuid.uuid4(),
                started=now,
                finished=now,
                status=const.ResourceState.undefined,
                action=const.ResourceAction.deploy,
                changes={},
                messages=[],
                change=const.Change.nochange,
                send_events=False,
                connection=connection,
            )

            skippable = await model.get_skipped_for_undeployable()
            if skippable:
                skippable_ids = [ResourceVersionIdStr(rid + ",v=%s" % version_id) for rid in skippable]
                # not checking error conditions
                await self.resource_service.resource_action_update(
                    env,
                    skippable_ids,
                    action_id=uuid.uuid4(),
                    started=now,
                    finished=now,
                    status=const.ResourceState.skipped_for_undefined,
                    action=const.ResourceAction.deploy,
                    changes={},
                    messages=[],
                    change=const.Change.nochange,
                    send_events=False,
                    connection=connection,
                )

        increments: tuple[abc.Set[ResourceIdStr], abc.Set[ResourceIdStr]] = await self.resource_service.get_increment(
            env, version_id
        )

        increment_ids, neg_increment = increments
        await self.resource_service.mark_deployed(env, neg_increment, now, version_id)

        # Setting the model's released field to True is the trigger for the agents to start pulling in the resources.
        # This has to be done after the resources outside of the increment have been marked as deployed.
        await model.update_fields(released=True, result=const.VersionState.deploying, connection=connection)

        if model.total == 0:
            await model.mark_done(connection=connection)
            return 200, {"model": model}

        if push:
            # fetch all resource in this cm and create a list of distinct agents
            agents = await data.ConfigurationModel.get_agents(env.id, version_id, connection=connection)
            await self.autostarted_agent_manager._ensure_agents(env, agents, connection=connection)

            for agent in agents:
                client = self.agentmanager_service.get_agent_client(env.id, agent)
                if client is not None:
                    if not agent_trigger_method:
                        env_agent_trigger_method = await env.get(ENVIRONMENT_AGENT_TRIGGER_METHOD, connection=connection)
                        incremental_deploy = env_agent_trigger_method == const.AgentTriggerMethod.push_incremental_deploy
                    else:
                        incremental_deploy = agent_trigger_method is const.AgentTriggerMethod.push_incremental_deploy
                    self.add_background_task(client.trigger(env.id, agent, incremental_deploy))
                else:
                    LOGGER.warning("Agent %s from model %s in env %s is not available for a deploy", agent, version_id, env.id)

        return 200, {"model": model}

    @handle(methods.deploy, env="tid")
    async def deploy(
        self,
        env: data.Environment,
        agent_trigger_method: const.AgentTriggerMethod = const.AgentTriggerMethod.push_full_deploy,
        agents: Optional[List[str]] = None,
    ) -> Apireturn:
        warnings = []

        # get latest version
        version_id = await data.ConfigurationModel.get_version_nr_latest_version(env.id)
        if version_id is None:
            return 404, {"message": "No version available"}

        # filter agents
        allagents = await data.ConfigurationModel.get_agents(env.id, version_id)
        if agents is not None:
            required = set(agents)
            present = set(allagents)
            allagents = list(required.intersection(present))
            notfound = required - present
            if notfound:
                warnings.append(
                    "Model version %d does not contain agents named [%s]" % (version_id, ",".join(sorted(list(notfound))))
                )

        if not allagents:
            return attach_warnings(404, {"message": "No agent could be reached"}, warnings)

        present = set()
        absent = set()

        await self.autostarted_agent_manager._ensure_agents(env, allagents)

        for agent in allagents:
            client = self.agentmanager_service.get_agent_client(env.id, agent)
            if client is not None:
                incremental_deploy = agent_trigger_method is const.AgentTriggerMethod.push_incremental_deploy
                self.add_background_task(client.trigger(env.id, agent, incremental_deploy))
                present.add(agent)
            else:
                absent.add(agent)

        if absent:
            warnings.append("Could not reach agents named [%s]" % ",".join(sorted(list(absent))))

        if not present:
            return attach_warnings(404, {"message": "No agent could be reached"}, warnings)

        return attach_warnings(200, {"agents": sorted(list(present))}, warnings)

    @handle(methods_v2.list_desired_state_versions, env="tid")
    async def desired_state_version_list(
        self,
        env: data.Environment,
        limit: Optional[int] = None,
        start: Optional[int] = None,
        end: Optional[int] = None,
        filter: Optional[Dict[str, List[str]]] = None,
        sort: str = "version.desc",
    ) -> ReturnValue[List[DesiredStateVersion]]:
        try:
            return await DesiredStateVersionView(
                environment=env,
                limit=limit,
                filter=filter,
                sort=sort,
                start=start,
                end=end,
            ).execute()
        except (InvalidFilter, InvalidSort, data.InvalidQueryParameter, data.InvalidFieldNameException) as e:
            raise BadRequest(e.message) from e

    @handle(methods_v2.promote_desired_state_version, env="tid")
    async def promote_desired_state_version(
        self,
        env: data.Environment,
        version: int,
        trigger_method: Optional[PromoteTriggerMethod] = None,
    ) -> None:
        if trigger_method == PromoteTriggerMethod.push_incremental_deploy:
            push = True
            agent_trigger_method = const.AgentTriggerMethod.push_incremental_deploy
        elif trigger_method == PromoteTriggerMethod.push_full_deploy:
            push = True
            agent_trigger_method = const.AgentTriggerMethod.push_full_deploy
        elif trigger_method == PromoteTriggerMethod.no_push:
            push = False
            agent_trigger_method = None
        else:
            push = True
            agent_trigger_method = None

        status_code, result = await self.release_version(
            env, version_id=version, push=push, agent_trigger_method=agent_trigger_method
        )
        if status_code != 200:
            raise BaseHttpException(status_code, result["message"])

    @handle(methods_v2.get_diff_of_versions, env="tid")
    async def get_diff_of_versions(
        self,
        env: data.Environment,
        from_version: int,
        to_version: int,
    ) -> List[ResourceDiff]:
        await self._validate_version_parameters(env.id, from_version, to_version)

        from_version_resources = await data.Resource.get_list(environment=env.id, model=from_version)
        to_version_resources = await data.Resource.get_list(environment=env.id, model=to_version)

        from_state = diff.Version(self.convert_resources(from_version_resources))
        to_state = diff.Version(self.convert_resources(to_version_resources))

        version_diff = to_state.generate_diff(from_state)

        return version_diff

    def convert_resources(self, resources: List[data.Resource]) -> Dict[ResourceIdStr, diff.Resource]:
        return {res.resource_id: diff.Resource(resource_id=res.resource_id, attributes=res.attributes) for res in resources}

    async def _validate_version_parameters(self, env: uuid.UUID, first_version: int, other_version: int) -> None:
        if first_version >= other_version:
            raise BadRequest(
                f"Invalid version parameters: ({first_version}, {other_version}). "
                "The second version number should be strictly greater than the first"
            )
        await self._check_version_exists(env, first_version)
        await self._check_version_exists(env, other_version)

    async def _check_version_exists(self, env: uuid.UUID, version: int) -> None:
        version_object = await data.ConfigurationModel.get_version(env, version)
        if not version_object:
            raise NotFound(f"Version {version} not found")



"""
    Copyright 2019 Inmanta

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Contact: code@inmanta.com
"""
import logging
from time import perf_counter, time
from typing import Optional

from pyformance import gauge
from pyformance.meters import Gauge

from inmanta.reporter import InfluxReporter
from inmanta.server import SLICE_METRICS
from inmanta.server import config as opt
from inmanta.server import protocol

LOGGER = logging.getLogger(__name__)


class MetricsService(protocol.ServerSlice):
    """Slice managing metrics collector"""

    def __init__(self) -> None:
        super(MetricsService, self).__init__(SLICE_METRICS)
        self._influx_db_reporter: Optional[InfluxReporter] = None

    async def start(self) -> None:
        self.start_auto_benchmark()
        self.start_metric_reporters()
        await super().start()

    async def stop(self) -> None:
        await super().stop()
        self.stop_metric_reporters()

    def stop_metric_reporters(self) -> None:
        if self._influx_db_reporter:
            self._influx_db_reporter.stop()
            self._influx_db_reporter = None

    def start_metric_reporters(self) -> None:
        if opt.influxdb_host.get():
            self._influx_db_reporter = InfluxReporter(
                server=opt.influxdb_host.get(),
                port=opt.influxdb_port.get(),
                database=opt.influxdb_name.get(),
                username=opt.influxdb_username.get(),
                password=opt.influxdb_password.get(),
                reporting_interval=opt.influxdb_interval.get(),
                autocreate_database=True,
                tags=opt.influxdb_tags.get(),
            )
            self._influx_db_reporter.start()

    def start_auto_benchmark(self) -> None:
        """Add all auto benchmarking to pyformance"""
        gauge("self.spec.cpu", CPUMicroBenchMark())


class CachingCallbackGuage(Gauge):
    """
    A gauge that calculates a value on the fly and holds it for a specific time interval.

    Intended to keep load under control.

    Be aware that the callback has to be short, as it is called from the IOLoop main thread.
    """

    def __init__(self, interval: float = 1):
        """
        :param interval: time a cache entry is valid, in seconds
        """
        super().__init__()
        self.next_time: float = 0
        self.last_value: int = 0
        self.interval: float = interval

    def get_value(self) -> int:
        now = time()
        if now < self.next_time:
            return self.last_value
        self.last_value = self.callback()
        self.next_time = now + self.interval
        return self.last_value

    def callback(self) -> int:
        """
        Be aware that the callback has to be short, as it is called from the IOLoop main thread.
        :return: the value of the gauge
        """
        raise NotImplementedError()


class CPUMicroBenchMark(CachingCallbackGuage):
    def callback(self) -> int:
        """
        Determine baseline performance of the machine by running a short cpu benchmark

        :return: time required to perform a specific calculation, in ns, 100% cpu bound
        """
        start = perf_counter()
        # this value was chosen to be around 0.1ms on a reference machine
        # which is long enough to be meaningful, short enough to not be disturbing.
        # the value is cached for 1s, making this at most 0.1% of additional overhead
        self.factor(1100)
        end = perf_counter()
        result = int((end - start) * 1000000000)
        return result

    @staticmethod
    def factor(number: int) -> int:
        """A CPU intensive algorithm"""
        out = 0
        # for each potential factor i
        for i in range(2, number + 1):
            # if i is a factor of N, repeatedly divide it out
            while number % i == 0:
                out += i
                number = number / i

        # if biggest factor occurs only once, n > 1
        if number > 1:
            out += number

        return out
